"""
Create protocol classes for all python classes in the PF python scripting reference pdf file (by running this python module). See output module 'pf_class_protocols' which is used in powfacpy. If you run this module a new module is created ('pf_class_protocols_new'). You can then delete the old one and rename the new file. 

IMPORTANT: If you miss a class please add it to the 'missing_classes_that_are_not_in_scripting_reference' list below.


This module is not used by other modules of powfacpy and has the sole purpose of creating the protocols.
"""

missing_classes_that_are_not_in_scripting_reference = [
    "IntFolder",
    "ElmSite",
    "ComVstab",
    "ComQvcurves",
    "ComMod",
    "ElmSecctrl",
    "ElmLod",
    "ElmIac",
    "ElmLodlv",
    "ElmLodmv",
    "ElmSind",
    "ElmVsc",
    "ElmVscmono",
    "ElmZpu",
    "StaPll",
    "TypSym",
    "TypLod",
    "BlkSlot",
    "BlkSum",
    "BlkRef",
    "StaPqmea",
    "CimArchive",
    "ComCimtogrid",
    "Evtlod",
    "EvtOutage",
    "EvtParam",
    "EvtShc",
    "EvtSwitch",
    "EvtSym",
    "EvtGen",
    "EvtTap",
    "EvtStop",
    "EvtSave",
    "EvtMessage",
    "EvtTransfer",
    "EvtPresync",
    "EvtTrigger",
    "EvtStep",
    "IntScan",
    "IntFilvers",
    "IntTemplate",
]


import os
import keyword
import sys
import json

from icecream import ic

powfacpy_project_root_directory = os.path.dirname(os.path.abspath(__file__)).replace(
    r"\src\powfacpy\pf_classes", ""
)
sys.path.insert(0, powfacpy_project_root_directory + r"\src")
import powfacpy


class Empty:
    """This empty class is used to get the generic attributes that every python object has (e.g. using 'dir(class_instance)').
    For the PF class protocols, these attributes are not added to the protocol classes of the PowerFactory objects.
    """

    pass


class PFClassesProtocolGenerator(powfacpy.PFActiveProject):
    """Generate protocol classes for PF classes from the PF python scripting reference."""

    def __init__(self, app):
        super().__init__(app)
        self.indentation_increment: str = "  "
        self.general_methods = self.get_general_methods()

    def create_pf_class_protocols(self) -> None:
        """Main method of this class."""
        pf_class_names = pfcpg.get_all_class_names_from_scripting_reference()
        pf_class_names += missing_classes_that_are_not_in_scripting_reference
        pf_class_objects, pf_class_objects_not_found, project = pfcpg.get_pf_objects(
            pf_class_names
        )
        print(pf_class_objects_not_found)
        try:
            self.app.Hide()
            with open(
                ".\\src\\powfacpy\\pf_classes\\protocols_new.py", "w"
            ) as out_file:
                out_file.write(
                    '"""Protocol classes for (almost) all PowerFactory classes in the python scripting reference pdf file. Protocol classes are helpful for example for type hints where they can be used just like \'normal\' implementations of classes.\n"""\n\n'
                )
                out_file.write("from typing import Protocol\n\n")
                out_file.write(self.get_general_class_string())
                for pf_obj in pf_class_objects:
                    definition: str = pfcpg.get_class_definition(pf_obj)
                    out_file.write(definition)
                    data_attributes, methods = pfcpg.get_class_attribute_definitions(
                        pf_obj
                    )
                    out_file.write(data_attributes + "\n\n")
                    out_file.write(methods + "\n\n")
        finally:
            project.Deactivate()
            project.Delete()
            self.app.Show()

    def get_all_class_names_from_scripting_reference(self) -> list[str]:
        """Uses a .txt file to which the TOC of the python scripting reference pdf document was copied (includes the class names).
        Iterates all words from the TOC and checks if they are PF class names.

        Returns:
            list[str]: PF class names
        """
        pf_class_names = []
        with open(
            powfacpy_project_root_directory
            + r"\power_factory_objects\CopyOfObjectsInTableOfContentsOfScriptingReference.txt",
            "r",
        ) as in_file:
            for line in in_file:
                words = line.split()
                for word in words:
                    if self.is_pf_class(word):
                        pf_class_names.append(word)
        return pf_class_names

    def get_pf_objects(self, pf_class_names: list[str]) -> list:
        """Get instances of the PF classes in 'pf_class_names'.

        Creates a new project and tries to create each class in one of the folders in the project. Some objects can only be created inside certain folders. Hence, the procedure is done twice to ensure that the folders are present (i.e. create folders in first run, use those folders to create the specific objects in the second run).
        Finally, further objects that cannot be created in this way are added.

        Args:
            pf_class_names (list[str]): _description_

        Returns:
            pf_class_objects: list of powerfactory class instances/objects
            pf_class_objects_not_found: list of objects that could not be found/created
        """
        try:
            self.app.Hide()
            project = self.app.CreateProject("delete_this_project", "grid")
            project.Activate()

            for n in range(2):
                folders = self.get_obj("*", include_subfolders=True)
                pf_class_objects = []
                pf_class_objects_not_found = []
                for pf_class in pf_class_names:
                    for folder in folders:
                        pf_class_obj = None
                        try:
                            pf_class_obj = self.create_in_folder(
                                "new." + pf_class, folder
                            )
                            if pf_class_obj:
                                break  # break loop when successfully created an object
                        except RuntimeError:  # CreateObject can raise runtime error
                            pass
                    if not pf_class_obj:
                        pf_class_objects_not_found.append(pf_class)
                    else:
                        pf_class_objects.append(pf_class_obj)
            # Insert some further objects
            pf_class_objects.insert(0, self.app)
            pf_class_objects.append(project)
            pf_class_objects_not_found.remove(pf_class_objects[-1].GetClassName())
            pf_class_objects.append(self.get_active_user_folder())
            pf_class_objects_not_found.remove(pf_class_objects[-1].GetClassName())
            pf_class_objects.append(self.get_first_level_folder("global library"))
            pf_class_objects_not_found.remove(pf_class_objects[-1].GetClassName())
        finally:
            self.app.Show()
        return pf_class_objects, pf_class_objects_not_found, project

    def get_class_definition(self, pf_obj) -> str:
        """Get string with the definition of a protocol class for a PF object.
        Example: 'class ElmTr2(Protocol):'

        Args:
            pf_obj: PF object

        Returns:
            str: class definition
        """
        definition: str = ""
        if not isinstance(pf_obj, type(self.app)):
            class_name = pf_obj.GetClassName()
            definition += "class " + class_name + "(PFGeneral):\n"
        else:
            class_name = "PFApp"
            definition += "class " + class_name + "(Protocol):\n"

        if hasattr(pf_obj, "GetClassDescription"):
            class_description = self.app.GetClassDescription(
                class_name
            )  # It seems that GetClassDescription always retuns an empty string for every class
            if class_description:
                definition += (
                    self.indentation_increment + '"""' + class_description + '"""\n\n'
                )
        return definition

    def get_class_attribute_definitions(self, pf_obj) -> str:
        """Get definitions of class attributes (data attributes and methods) of a PF object.

        Args:
            pf_obj: PF object

        Returns:
            str: definitions of all attributes of the class of the object
        """
        class_attributes: set = (
            set(dir(pf_obj)) - set(dir(Empty)) - self.general_methods
        )  # omit attributes that all python objects have and general methods.
        class_attributes = sorted(class_attributes)
        data_attributes = ""
        methods = ""
        for attr in class_attributes:
            # In rare cases, the attributes are named after python keywords - such names are not allowed in python classes and are omitted.
            if hasattr(pf_obj, attr) and not keyword.iskeyword(attr):
                if callable(getattr(pf_obj, attr)):
                    # Due to lack of knowledge about the method arguments, the generic '(*args)' is used.
                    methods += (
                        self.indentation_increment
                        + "def "
                        + attr
                        + "(*args):\n"
                        + 2 * self.indentation_increment
                        + "...\n\n"
                    )
                else:
                    attribute_value = getattr(pf_obj, attr)
                    attribute_type = type(attribute_value).__name__
                    # The types "DataObject", "NoneType" are replace with 'object'
                    # "DataObject" is replaced because powfacpy should not depend on an import of the powerfactory module (which is necessary for this type hint)
                    # "NoneType" is not a helpful information
                    if not attribute_type in ["DataObject", "NoneType"]:
                        data_attributes += (
                            self.indentation_increment
                            + attr
                            + ": "
                            + attribute_type
                            + "\n"
                        )
                    else:
                        data_attributes += (
                            self.indentation_increment + attr + ": object\n"
                        )
                    if not isinstance(pf_obj, type(self.app)):
                        attribute_description = pf_obj.GetAttributeDescription(attr)
                        if attribute_description:
                            # some characters need to be deleted (newlines, '\xb0' is not a unicode symbol)
                            attribute_description = (
                                attribute_description.replace("\n", "")
                                .replace('"', "")
                                .replace("\xb0", "")
                            )
                            data_attributes += (
                                self.indentation_increment
                                + '"'
                                + attribute_description
                                + '"\n'
                            )
        return data_attributes, methods

    def get_general_class_string(self):
        """Add class with 'general methods' (see Section 'General Methods' in scripting reference). All other classes (except for PFApp) inherit from this class."""
        class_str = f'class PFGeneral(Protocol):\n{self.indentation_increment}"""Class with general methods (see Section \'General Methods\' in scripting reference. All other classes (except for PFApp) inherit from this class). \n{self.indentation_increment}"""\n'
        for method in self.get_general_methods():
            class_str += (
                self.indentation_increment
                + "def "
                + method
                + "(*args):\n"
                + 2 * self.indentation_increment
                + "...\n\n"
            )
        return class_str

    def get_general_methods(self):
        """General methods from python scripting reference."""
        general_methods = "AddCopy ContainsNonAsciiCharacters CopyData CreateObject Delete Energize GetAttribute GetAttributeDescription GetAttributeLength GetAttributes GetAttributeShape GetAttributeType GetAttributeUnit GetChildren GetClassName GetCombinedProjectSource GetConnectedElements GetConnectionCount GetContents GetControlledNode GetCubicle GetFullName GetImpedance GetInom GetNode GetOperator GetOwner GetParent GetReferences GetRegion GetSupplyingSubstations GetSupplyingTransformers GetSupplyingTrfstations GetSystemGrounding GetUnom GetZeroImpedance HasAttribute HasResults IsCalcRelevant IsDeleted IsEarthed IsEnergized IsHidden IsInFeeder IsNetworkDataFolder IsNode IsObjectActive IsObjectModifiedByVariation Isolate IsOutOfService IsReducible IsShortCircuited MarkInGraphics Move PasteCopy PurgeUnusedObjects ReplaceNonAsciiCharacters ReportNonAsciiCharacters ReportUnusedObjects SearchObject SetAttribute SetAttributeLength SetAttributes SetAttributeShape ShowEditDialog ShowModalSelectTree SwitchOff SwitchOn WriteChangesToDb"
        return set(general_methods.split())


if __name__ == "__main__":
    with open(powfacpy_project_root_directory + "\\settings.json") as settings_file:
        settings = json.load(settings_file)
    sys.path.append(settings["local path to PowerFactory application"])
    import powerfactory

    app = powerfactory.GetApplication()
    pfcpg = PFClassesProtocolGenerator(app)
    pfcpg.create_pf_class_protocols()



from __future__ import annotations

from icecream import ic

from powfacpy.pf_classes.protocols import PFGeneral
from powfacpy.applications.caching import (
    cache_attr,
    cache_method,
)


class BaseObjectStatic:
    """Base class that can be used to store a PF object (in attribute '_obj'). Adds functionality for checking equality and string representation. Uses slots for efficient attribute access and memory usage (read more about slots in the documentation: https://wiki.python.org/moin/UsingSlots)."""

    __slots__ = "_obj"

    def __init__(self, obj: PFGeneral) -> None:
        self._obj: PFGeneral = obj

    def __eq__(self, value: BaseObjectStatic) -> bool:
        """Check equality with a PF object.

        Args:
            value (BaseObjectStatic)

        Returns:
            bool: If equal
        """
        return self._obj == value._obj

    def __str__(self):
        """Returns the 'loc_name' of the PF object."""
        return self._obj.loc_name

    @property
    def obj(self):
        return self._obj


class BaseChildStatic(BaseObjectStatic):
    """
    Base child class (in object oriented sense - not like in GetParent()) to be used in combination with instances of PF objects (stored in '_obj'). The '__getattribute__' and '__setattr__' methods are overloaded to resemble the behavior of inheriting from the PF object. That means that all methods/data attributes of '_obj' are available and further attributes/methods can be added by any class that inherits from 'BaseChildStatic'.

    In addition, functionality is added to cache attributes and methods (the data is stored in '_obj' because 'BaseChildStatic' is static, i.e. it has no '__dict__' attribute because of the static slots).

    '__slots__' is required to inherit the slots from 'BaseObjectStatic' (i.e. '__slots__ = "_obj"').
    """

    __slots__ = ()

    def __init__(self, obj: PFGeneral) -> None:
        super(BaseChildStatic, self).__setattr__("_obj", obj)

    def __getattribute__(self, attr):
        _obj = super(BaseChildStatic, self).__getattribute__("_obj")
        if attr == "_obj":
            return _obj
        try:
            return _obj.__getattribute__(attr)
        except AttributeError:
            try:
                return _obj.__getattr__(attr)
            except AttributeError:  # Fallback that should never happen
                return super(BaseChildStatic, self).__getattribute__(attr)

    def __setattr__(self, attr, value):
        if hasattr(self._obj, attr):
            self._obj.__setattr__(attr, value)
        else:
            super(BaseChildStatic, self).__setattr__(attr, value)

    def cache_attr(self, attr: str) -> None:
        cache_attr(self._obj, attr)

    def cache_method(self, method: str):
        cache_method(self._obj, method)


class BaseChild(BaseChildStatic):
    """This is a dynamic alternative to 'BaseChildStatic', i.e. just by declaring the class (without '__slots__ = ()', a '__dict__' is added and attributes can be added dynamical by any class that inherits from 'BaseChild'"""



import sys
import importlib

import pytest

sys.path.insert(0, r".\src")
import powfacpy
from powfacpy.base.base import BaseObjectStatic, BaseChild, BaseChildStatic
from powfacpy.base.active_project import ActiveProject

importlib.reload(powfacpy)


def test_base_obj_static(act_prj: ActiveProject, activate_powfacpy_test_project):
    terminal = act_prj.get_calc_relevant_obj("*.ElmTerm")[0]
    terminal_base_obj_1 = BaseObjectStatic(terminal)
    terminal_base_obj_2 = BaseObjectStatic(terminal)
    assert terminal_base_obj_1.obj == terminal_base_obj_2._obj  # test property obj
    assert terminal_base_obj_1 == terminal_base_obj_2  # test __eq__
    assert str(terminal_base_obj_1) == terminal.loc_name  # test __str__


def test_base_child_caching(act_prj: ActiveProject, activate_powfacpy_test_project):
    terminal = act_prj.get_calc_relevant_obj("*.ElmTerm")[0]
    terminal_base_child = BaseChild(terminal)
    terminal_base_child.cache_attr("uknom")
    assert terminal_base_child.__dict__["uknom"] == terminal.uknom

    terminal_base_child.cache_method("GetClassName")
    assert "GetClassName" in terminal_base_child.__dict__
    assert terminal_base_child.GetClassName() == terminal.GetClassName()


def test_base_child_static(act_prj: ActiveProject, activate_powfacpy_test_project):
    terminal = act_prj.get_calc_relevant_obj("*.ElmTerm")[0]
    terminal_base_child_static = BaseChildStatic(terminal)
    # Call methods, get attributes..
    cub = terminal_base_child_static.CreateObject("StaCubic", "a")
    terminal_base_child_static.AddCopy(cub)
    assert terminal_base_child_static.GetParent() == terminal.GetParent()
    assert terminal_base_child_static.GetContents() == terminal.GetContents()
    assert terminal_base_child_static.uknom == terminal.uknom

    # cannot dynamically add attributes
    with pytest.raises(AttributeError):
        terminal_base_child_static.foo


if __name__ == "__main__":
    # pytest.main([r"tests\test_base.py"])
    # pytest.main([r"tests\applications\test_networks.py"])
    pytest.main([r"tests\applications\test_model_exchange.py"])



import sys

import pytest
import numpy as np
import pandas as pd

sys.path.insert(0, r".\src")
from powfacpy.pf_classes.elm.file import MeasurementFile
from powfacpy.base.active_project import ActiveProject
from powfacpy.pf_classes.protocols import IntPrj


@pytest.fixture(
    scope="function",
)
def create_measurement_file(
    act_prj,
) -> MeasurementFile:
    # activate_powfacpy_test_project.Activate()
    # act_prj = ActiveProject()
    # act_prj.app.ActivateProject(
    #     r"\seberlein\powfacpy\powfacpy_tests_copy_where_tests_run"
    # )
    print(f"############# {act_prj.app.GetActiveProject()}")
    elmfile = act_prj.create_in_folder(
        "Measurement file.ElmFile",
        folder=r"Network Model\Network Data\test_dyn_sim_interface\Grid 1\test_composite_model",
    )
    return MeasurementFile(elmfile)


def test_from_pandas_dataframe(
    activate_powfacpy_test_project,
    create_measurement_file: MeasurementFile,
):
    meas_file = create_measurement_file
    df = pd.DataFrame(np.array([[1, 2, 3], [4, 5, 6]]), index=[0, 1])
    meas_file.from_pandas_dataframe(df, "test")


def test_get_full_path_of_elmfile_data(activate_powfacpy_test_project):
    MeasurementFile.get_full_path_of_elmfile_data_in_external_data_directory()
    MeasurementFile.get_path_of_elmfile_data_inside_external_data_directory()


if __name__ == "__main__":
    # pytest.main([r"tests\pf_classes\elm\test_file.py"])
    # pytest.main([r"tests\pf_classes"])
    # pytest.main([r"tests", r"--ignore=tests\deprecated"])
    pytest.main([r"tests"])



import sys

import pytest

sys.path.insert(0, r".\src")
from powfacpy.pf_classes.elm.boundary import Boundary
from powfacpy.base.active_project import ActiveProject
from powfacpy.applications.topology import Topology
from powfacpy.pf_classes.protocols import PFGeneral
from powfacpy.result_variables import ResVar

RMS_BAL = ResVar.RMS_Bal


@pytest.fixture(scope="function")
def create_boundary(
    act_prj: ActiveProject, activate_39_bus_new_england_test_project
) -> tuple[Boundary, list[PFGeneral]]:
    topo = Topology()
    terminals = act_prj.get_calc_relevant_obj(
        "Bus 36.ElmTerm, Bus 23.ElmTerm, Bus 22.ElmTerm, Bus 35.ElmTerm"
    )
    boundary = topo.create_boundary_without_changing_initial_zones(
        "South-East", terminals
    )
    return boundary, terminals


def test_get_all_internal_elms(create_boundary):
    boundary, terminals = create_boundary
    boundary = Boundary(boundary)
    internal_terminals = boundary.get_internal_elms(
        condition=lambda x: x.GetClassName() == "ElmTerm"
    )
    assert len(internal_terminals) == len(terminals)


def test_add_results_variable_for_elms(create_boundary):
    boundary, terminals = create_boundary
    boundary = Boundary(boundary)
    boundary.add_results_variable_for_elms(
        lambda x: x.GetClassName() == "ElmTerm",
        [RMS_BAL.ElmTerm.m_fe.value, RMS_BAL.ElmTerm.m_u.value],
    )


def test_caching(create_boundary):
    boundary, terminals = create_boundary
    boundary = Boundary(boundary)
    boundary.cache_method("GetInterior")
    assert boundary.GetInterior() == boundary._obj.GetInterior()


if __name__ == "__main__":
    pytest.main([r"tests\pf_classes\elm\test_boundary.py"])



"Interface to pandas package."

import pandas as pd

from powfacpy.pf_classes.protocols import PFApp
from powfacpy.applications.application_base import ApplicationBase
from powfacpy.pf_classes.protocols import (
    PFApp,
)


class PandasInterface(ApplicationBase):

    def __init__(
        self, pf_app: PFApp | None | bool = False, cached: bool = False
    ) -> None:
        super().__init__(pf_app, cached)

    def replace_loc_name_with_pf_objects_in_labels(
        self,
        df: pd.DataFrame,
        class_name: str,
        only_calc_relevant: bool = True,
        index_and_column_labels_are_equal: bool = False,
        level: int = 0,
    ) -> pd.DataFrame:
        """Replaces string labels ('loc_name' of PF objects) of a dataframe with the corresponding PF objects.

        It is assumed that all PF objects are of the same class.

        Args:
            df (pd.DataFrame): index and column labels are 'loc_name' attributes of PF objects

            class_name (str): PF class name of the labels (e.g. 'ElmTerm'). Note that all PF objects must be of the same class.

            only_calc_relevant (bool): Search only for calculation relevant objects (using 'GetCalcRelevantObjects'). If False, search all objects (using 'get_obj') Defaults to True.

            index_and_column_labels_are_equal (bool, optional): If True, it is assumed that indices and columns have the same labels (improved performance). This is common for symmetric matrices. Defaults to False.

            level (int, optional): If labels are MultiIndex, the level that is replaced can be specified. Defaults to 0.

        Returns:
            pd.DataFrame: Frame with updated labels
        """
        separator = f".{class_name},"
        all_objs = separator.join(df.columns.get_level_values(level).to_list())
        if only_calc_relevant:
            all_objs = self.app.GetCalcRelevantObjects(all_objs)
        else:
            all_objs = self.act_prj.get_obj(all_objs, include_subfolders=True)
        df.columns = all_objs
        if not index_and_column_labels_are_equal:
            all_objs = separator.join(df.index.get_level_values(level).to_list())
            all_objs = self.app.GetCalcRelevantObjects(all_objs)
        df.index = all_objs
        return df



import sys
import os
import json
import pytest

with open(".\\settings.json") as settings_file:
    settings = json.load(settings_file)
sys.path.append(settings["local path to PowerFactory application"])
import powerfactory

sys.path.insert(0, r".\src")
from powfacpy.base.active_project import ActiveProject
from powfacpy.pf_classes.protocols import IntPrj, PFApp


@pytest.fixture(scope="session")
def pf_app():
    if settings["PowerFactory username"]:
        if settings["PowerFactory command line arguments for GetApplication"]:
            return powerfactory.GetApplicationExt(
                settings["PowerFactory username"],
                settings["PowerFactory password"],
                settings["PowerFactory command line arguments for GetApplication"],
            )
        else:
            return powerfactory.GetApplicationExt(
                settings["PowerFactory username"],
                settings["PowerFactory password"],
            )
    else:
        return powerfactory.GetApplicationExt()


@pytest.fixture(scope="session")
def act_prj(pf_app: PFApp) -> ActiveProject:
    # Return ActiveProject instance
    return ActiveProject(pf_app)


@pytest.fixture(scope="module")
def copy_test_project(act_prj: ActiveProject) -> IntPrj:
    """This method will create a copy of that project which is then used for the tests. This ensures that the tests are always run with the same initial
    project state.
    """
    return create_copy_of_test_project(act_prj, "powfacpy_tests")


@pytest.fixture(scope="module")
def copy_39_bus_new_england_test_project(act_prj: ActiveProject) -> IntPrj:
    """This method will create a copy of that project which is then used for the tests. This ensures that the tests are always run with the same initial
    project state.
    """
    return create_copy_of_test_project(act_prj, "39_bus_new_england")


def create_copy_of_test_project(
    act_prj: ActiveProject, project_name_in_powfacpy_folder: str
) -> IntPrj:
    user = act_prj.app.GetCurrentUser()
    powfacpy_folder_path = settings["path to powfacpy folder in PowerFactory database"]
    project_for_testing = act_prj.get_unique_obj(
        rf"{powfacpy_folder_path}\{project_name_in_powfacpy_folder}",
        parent_folder=user,
        include_subfolders=False,
    )
    folder_of_project_for_testing = act_prj.get_unique_obj(
        powfacpy_folder_path, parent_folder=user, include_subfolders=False
    )
    project_copy = act_prj.copy_single_obj(
        project_for_testing,
        folder_of_project_for_testing,
        new_name=f"{project_name_in_powfacpy_folder}_copy_where_tests_run",
    )
    return project_copy


@pytest.fixture(scope="function")
def activate_powfacpy_test_project(
    copy_test_project: IntPrj, act_prj: ActiveProject
) -> IntPrj:
    """The project for testing must be located in the current
    user under "powfacpy\\powfacpy_tests".
    """
    # ap = act_prj.app.GetActiveProject()
    # if ap is None:
    copy_test_project.Activate()
    # if not ap == copy_test_project:
    #     ap.Deactivate()
    #     copy_test_project.Activate()
    # assert act_prj.app.GetActiveProject() == copy_test_project
    return copy_test_project


@pytest.fixture(scope="function")
def activate_39_bus_new_england_test_project(
    copy_39_bus_new_england_test_project: IntPrj,
) -> IntPrj:
    """The project for testing must be located in the current
    user under "powfacpy\\powfacpy_tests".
    """
    copy_39_bus_new_england_test_project.Activate()
    return copy_39_bus_new_england_test_project



import sys
from os import remove, getcwd

import pytest

sys.path.insert(0, r".\src")
import powfacpy
from powfacpy.applications.dynamic_simulation import DynamicSimulation
from powfacpy.applications.results import Results
from powfacpy.pf_classes.protocols import PFApp
import importlib

importlib.reload(powfacpy)


@pytest.fixture
def pfsim(pf_app: PFApp) -> DynamicSimulation:
    # Return PFDynSimInterface instance
    return DynamicSimulation(pf_app)


def test_export_to_csv(pfsim: DynamicSimulation, activate_powfacpy_test_project):
    export_dir = getcwd() + r"\tests\tests_output"
    file_name = "test_results"
    study_case = pfsim.act_prj.get_unique_obj(
        r"Study Cases\test_dyn_sim_interface\Study Case"
    )
    study_case.Activate()
    pfsim.act_prj.add_results_variable(
        r"Network Model\Network Data\test_dyn_sim_interface\Grid 1\AC Voltage Source",
        "m:Psum:bus1",
    )
    pfsim.initialize_and_run_sim()

    pfri = Results(pfsim.act_prj.app)
    export_dir = export_dir
    pfri.export_to_csv(export_dir)
    remove(export_dir + "\\results.csv")

    pfri.export_to_csv(dir=export_dir, file_name=file_name)
    remove(export_dir + "\\" + file_name + ".csv")

    results_obj = pfsim.act_prj.get_unique_obj(
        r"Study Cases\test_dyn_sim_interface\Study Case\All calculations"
    )
    pfri.export_to_csv(export_dir, results_obj=results_obj)
    remove(export_dir + "\\results.csv")


def test_set_and_get_dsl_obj_array(
    pfsim: DynamicSimulation, activate_powfacpy_test_project
):
    array = [[2, 0, 2, 0], [1, 2, 3, 4], [5, 6, 7, 8]]
    array_two_column = [[2, 0], [1, 2], [5, 6]]
    dsl_obj = pfsim.act_prj.get_unique_obj(
        r"Network Model\Network Data\test_dyn_sim_interface\Grid 1\Voltage source ctrl\Angle"
    )

    pfsim.set_dsl_obj_array(dsl_obj, array)
    array_returned = powfacpy.PFDynSimInterface.get_dsl_obj_array(dsl_obj)
    assert array_returned == array

    pfsim.set_dsl_obj_array(dsl_obj, array, size_included_in_array=False)
    array_returned = pfsim.get_dsl_obj_array(dsl_obj, size_included_in_array=False)
    assert array_returned == array

    pfsim.set_dsl_obj_array(dsl_obj, array_two_column, array_num=2)
    array_returned = pfsim.get_dsl_obj_array(dsl_obj, array_num=2)
    assert array_returned == array_two_column

    pfsim.set_dsl_obj_array(
        dsl_obj, array_two_column, array_num=2, size_included_in_array=False
    )
    array_returned = pfsim.get_dsl_obj_array(
        dsl_obj, array_num=2, size_included_in_array=False
    )
    assert array_returned == array_two_column


def test_create_event(pfsim: DynamicSimulation, activate_powfacpy_test_project):

    target = pfsim.act_prj.get_unique_obj(
        r"Network Model\Network Data\test_dyn_sim_interface\Grid 1\AC Voltage Source"
    )
    pfsim.create_dyn_sim_event(
        "test.EvtParam",
        {"time": 1, "p_target": target, "variable": "u0", "value": "1.05"},
    )

    target = pfsim.act_prj.get_unique_obj(
        r"Network Model\Network Data\test_dyn_sim_interface\Grid 1\General Load HV"
    )
    pfsim.create_dyn_sim_event(
        "loadevent.EvtLod", {"time": 1, "p_target": target, "dP": 100}
    )


def test_get_parameters_of_dsl_models_in_composite_model(
    pfsim: DynamicSimulation, activate_powfacpy_test_project
):
    pfsim.act_prj.get_unique_obj(
        r"Study Cases\test_dyn_sim_interface\Study Case"
    ).Activate()
    composite_model = pfsim.act_prj.get_unique_obj(
        r"Network Model\Network Data\test_dyn_sim_interface\Grid 1\test_composite_model"
    )

    par_val_dict_1 = pfsim.get_parameters_of_dsl_models_in_composite_model(
        composite_model
    )
    assert not par_val_dict_1["test_no_parameters"]
    assert par_val_dict_1["test_a"]["d"] == 0
    assert not "oarray_x" in par_val_dict_1["test_a"]

    par_val_dict_2 = pfsim.get_parameters_of_dsl_models_in_composite_model(
        composite_model, single_dict_for_all_dsl_models=True
    )
    assert par_val_dict_2["d"] == 0

    # Raise exception if parameter values are not consistent for all DSL models
    # and a single dict is used.
    par_val_dict_1["test_a"]["a"] = 199
    pfsim.set_parameters_of_dsl_models_in_composite_model(
        composite_model, par_val_dict_1
    )
    with pytest.raises(
        powfacpy.exceptions.PFInconsistentParamValueOfDSLModelInCompositeModel
    ):
        pfsim.get_parameters_of_dsl_models_in_composite_model(
            composite_model, single_dict_for_all_dsl_models=True
        )


def test_set_parameters_of_dsl_models_in_composite_model(
    pfsim: DynamicSimulation, activate_powfacpy_test_project
):
    pfsim.act_prj.get_unique_obj(
        r"Study Cases\test_dyn_sim_interface\Study Case"
    ).Activate()
    composite_model = pfsim.act_prj.get_unique_obj(
        r"Network Model\Network Data\test_dyn_sim_interface\Grid 1\test_composite_model"
    )

    par_val_dict = {"test_a": {"a": 50}}
    pfsim.set_parameters_of_dsl_models_in_composite_model(composite_model, par_val_dict)
    par_val_dict = pfsim.get_parameters_of_dsl_models_in_composite_model(
        composite_model
    )
    assert par_val_dict["test_a"]["a"] == 50

    par_val_dict = {"b": 70}
    pfsim.set_parameters_of_dsl_models_in_composite_model(
        composite_model, par_val_dict, single_dict_for_all_dsl_models=True
    )
    par_val_dict = pfsim.get_parameters_of_dsl_models_in_composite_model(
        composite_model
    )
    assert par_val_dict["test_a"]["b"] == 70
    assert par_val_dict["test_b"]["b"] == 70

    # Make sure there is no exception when par_val_dict is empty
    par_val_dict = {}
    pfsim.set_parameters_of_dsl_models_in_composite_model(
        composite_model, par_val_dict, single_dict_for_all_dsl_models=True
    )


def test_get_dsl_model_parameter_names(
    pfsim: DynamicSimulation, activate_powfacpy_test_project
):
    pfsim.act_prj.get_unique_obj(
        r"Study Cases\test_dyn_sim_interface\Study Case"
    ).Activate()
    composite_model = pfsim.act_prj.get_unique_obj(
        r"Network Model\Network Data\test_dyn_sim_interface\Grid 1\test_composite_model_no_blkdef"
    )
    dsl_model = pfsim.act_prj.get_unique_obj("test_a", parent_folder=composite_model)
    with pytest.raises(powfacpy.exceptions.PFObjectAttributeTypeError):
        pfsim.get_dsl_model_parameter_names(dsl_model)


if __name__ == "__main__":
    pytest.main(([r"tests\applications\test_dynamic_simulation.py"]))



from __future__ import annotations

from powfacpy.pf_classes.protocols import SetColscheme, PFGeneral
from powfacpy.pf_classes.elm.elm_base import ElmBase
from powfacpy.pf_classes.elm.grouping_base import GroupingBase
from powfacpy.result_variables import ResVar

RMS_BAL = ResVar.RMS_Bal
LF_BAL = ResVar.LF_Bal


class DiagramColorScheme(ElmBase):

    __slots__ = ()

    def __init__(self, obj: SetColscheme) -> None:
        super().__init__(obj)
        self._obj: SetColscheme

    def __new__(cls, *args, **kwargs) -> SetColscheme | DiagramColorScheme:
        """Implemented only to add type hints for the instance.

        Returns:
            SetColscheme | DiagramColorScheme: New instance
        """
        instance = super().__new__(cls)
        return instance

    def show_boundary_interior_regions(self) -> None:
        self._obj.cUseColouring = 1
        self._obj.cGroup = 1
        self._obj.cColouring = 20



from __future__ import annotations

from powfacpy.base.active_project import ActiveProjectCached
from powfacpy.pf_classes.protocols import StaPll

from powfacpy.pf_classes.elm.elm_base import ElmBase
from powfacpy.result_variables import ResVar

RMS_BAL = ResVar.RMS_Bal


class PhaseLockedLoop(ElmBase):

    __slots__ = ()

    def __init__(self, obj: StaPll) -> None:
        super().__init__(obj)
        self._obj: StaPll

    def __new__(cls, *args, **kwargs) -> StaPll | PhaseLockedLoop:
        """Implemented only to add type hints for the created instance.

        Returns:
            StaPll | PhaseLockedLoop: New instance
        """
        instance = super().__new__(cls)
        return instance

    def set_standard_parameters(self):
        self._obj.i_norm = 1  # improved behavior during large voltage deviations



import powfacpy
from powfacpy.pf_class_protocols import PFApp


class PFStringManipulation:
    """Class to manipulate strings in the PF context."""

    @staticmethod
    def replace_between_characters(
        char1: str, char2: str, replacement: str, original: str
    ):
        """Replace between 'char1' and 'char2' in 'original' with 'replacement'.

        Example:
          Calling
            powfacpy.PFStringManipulation.replace_between_characters(
              '.',
              '\\',
              '\\',
              'username.IntUser\\pow.facpy.\\powfacpy.tests.IntPrj\\Network Model.IntPrjfolder\\Network Data.IntPrjfolder\\test_base_interface\\Grid.ElmNet\\Terminal HV 1.ElmTerm'
          would give the output:
            'username\\pow.facpy\\powfacpy.tests\\Network Model\\Network Data\\test_base_interface\\Grid\\Terminal HV 1'
          Note the behavior when there are several '.' in between '\\'
          -> then the replacement starts after the last '.'
        """
        new_string = ""
        is_after_char_1 = False
        string_between_char_1_occurrences = ""
        for c in original:
            if c == char1:
                is_after_char_1 = True
                new_string += string_between_char_1_occurrences
                string_between_char_1_occurrences = ""
            elif c == char2:
                if is_after_char_1:
                    new_string += replacement
                else:
                    new_string += c
                string_between_char_1_occurrences = ""
                is_after_char_1 = False
            if is_after_char_1:
                string_between_char_1_occurrences += c
            elif not c == char2:
                new_string += c
        return new_string

    @staticmethod
    def remove_html_tags_from_path(path: str) -> str:
        return path[path.find(">") + 1 : path.rfind("<")]

    @staticmethod
    def remove_opening_html_tag_from_path(path: str) -> str:
        return path[path.find(">") + 1 :]

    @staticmethod
    def remove_closing_html_tag_from_path(path: str) -> str:
        return path[0 : path.rfind("<")]

    @staticmethod
    def remove_class_names(path: str) -> str:
        return PFStringManipulation.replace_between_characters(".", "\\", "\\", path)

    @staticmethod
    def truncate_until(original: str, string_pattern: str) -> str:
        """Truncate all characters until (and including) the occurence of string_pattern in original."""
        return original[original.find(string_pattern) + len(string_pattern) :]

    @staticmethod
    def truncate_beginning(original: str, string_pattern: str) -> str:
        """Truncate string_pattern if it occurs at the beginning of the
        original string. Otherwise return original string.
        """
        index_where_pattern_begins = original.find(string_pattern)
        if index_where_pattern_begins == 0:
            return original[len(string_pattern) + 1 :]
        else:
            return original

    @staticmethod
    def format_full_path(path: str, pf_app: PFApp) -> str:
        """Takes the full path (including user and project) and returns the path relative to the currently active project. Deletes class information.

        Examples:
            ```
            input path:  \\username.IntUser\\powfacpy_base.IntPrj\\Network Model.IntPrjfolder\\Network Data.IntPrjfolder\\Grid.ElmNet\\Terminal 1.ElmTerm
            output: Network Model\\Network Data\\Grid\\Terminal 1
            ```
        """
        path = PFStringManipulation._get_path_inside_active_project_from_full_path(
            path, pf_app
        )
        return PFStringManipulation.remove_class_names(path)

    @staticmethod
    def format_variable_name(name: str) -> str:
        """Takes PF-generated csv export variable name and returns shortened version.
        Example:
          name: 's:u0 in kV'
          output: 's:u0'
        """
        return (
            name.split(" ", 1)[0].replace('"', "").replace("\n", "")
        )  # get rid of description and quotation marks

    @staticmethod
    def handle_path(path: str):
        """Checks if path starts with '\\' (not accepted by most PF methods)
        and if 'path' is of type string.
        """
        try:
            if not path[0] == "\\":
                return path
            else:
                return path[1:]
        except TypeError:
            raise TypeError("Path must be of type string.")

    @staticmethod
    def replace_outside_or_inside_of_strings_in_a_string(
        string: str, replacements: dict, outside=True
    ):
        """This method replaces parts of a string but only in the sections
        of the original string that are either outside of strings.

        Example:
        "p HV load >= 2 and 'This is a string inside the string'"
        In this string the part 'This is a string inside the string'
        is a string inside the original string and no replacements are
        made in this part.

        Args:
            string: The string that will be adjusted

            replacements: key-value pairs of matching strings and their replacement
        """
        if outside:
            modulo_result = 0
        else:
            modulo_result = 1
        string = PFStringManipulation.split_but_keep_delimiter(string, "'")
        for n, part_of_string in enumerate(string):
            if (
                n % 2 == modulo_result
            ):  # Even elemnts store the parts that are outside strings
                for par_name, var_name in replacements.items():
                    part_of_string = part_of_string.replace(par_name, var_name)
                    string[n] = part_of_string
        return "".join(string).strip()

    @staticmethod
    def split_but_keep_delimiter(string: str, delimiter: str):
        """Uses the split() method to separate a string acoording
        to a delimiter, but keeps the delimiter in the
        separated strings (it is suprising that this is not optional
        in the split() method).
        Example:
          split_but_keep_delimiter(
            "p HV load >= 2 and (control 1 == 'A' and control 2 != 'S')",
            delimiter: "'")
          returns
            ["p HV load >= 2 and (control 1 == '", "A'", " and control 2 != '", "S'", ')']
        """
        split_strings_list = [part + delimiter for part in string.split(delimiter)]
        split_strings_list[-1] = split_strings_list[-1].rstrip(delimiter)
        return split_strings_list

    @staticmethod
    def _get_path_inside_active_project_from_full_path(path: str, pf_app: PFApp) -> str:
        """
        Takes the full path (including user and project) and returns the path
        relative to the currently active project.
        Example:
          input path:  \\username.IntUser\\powfacpy_base.IntPrj\\Network Model.IntPrjfolder\\Network Data.IntPrjfolder\\Grid.ElmNet\\Terminal 1.ElmTerm</l3>
          output: Network Model.IntPrjfolder\\Network Data.IntPrjfolder\\Grid.ElmNet\\Terminal 1.ElmTerm
        """
        project_name = pf_app.GetActiveProject().loc_name + ".IntPrj\\"
        path = PFStringManipulation.truncate_until(path, project_name)
        # In case a closing tag occurs at the end of the path </l3> (e.g. when
        # str() is called on a PF object) make sure this is removed.
        if path and path[-1] == ">":
            path = path[0 : path.rfind("<")]
        return path


class PFTranslator:
    """Class to translate between languages.
    Currently not used anywhere in powfacpy any more.
    """

    @staticmethod
    def get_default_result_object_name(language):
        if language == "en":
            return "All calculations.ElmRes"
        elif language == "de":
            return "Alle Berechnungsarten.ElmRes"

    @staticmethod
    def get_default_graphics_board_name(language):
        if language == "en":
            return "Graphics Board.SetDesktop"
        elif language == "de":
            return "Grafiksammlung.SetDesktop"

    @staticmethod
    def get_default_study_case_folder_name(language):
        if language == "en":
            return "Study Cases.IntPrjfolder"
        elif language == "de":
            return "Berechnungsfälle.IntPrjfolder"

    @staticmethod
    def get_default_operation_scenario_folder_path(language):
        if language == "en":
            return r"Network Model\Operation Scenarios"
        elif language == "de":
            return r"Netzmodell\Betriebsfälle"

    @staticmethod
    def get_default_variations_folder_path(language):
        if language == "en":
            return r"Network Model\Variations"
        elif language == "de":
            return r"Netzmodell\Varianten"

    @staticmethod
    def _get_language_dependent_name_from_studycase(
        studycase, english_name, german_name
    ):
        studycase_contents = powfacpy.PFTranslator.get_name_with_ending(
            studycase.GetContents()
        )
        has_english_name = english_name in studycase_contents
        has_german_name = german_name in studycase_contents
        assert not (
            has_english_name and has_german_name
        ), "Two redundant file versions: English and German named files exist."
        if has_english_name:
            return english_name
        else:
            return german_name

    @staticmethod
    def get_graphics_board_name_from_studycase(studycase):
        name = powfacpy.PFTranslator._get_language_dependent_name_from_studycase(
            studycase=studycase,
            english_name="Graphics Board.SetDesktop",
            german_name="Grafiksammlung.SetDesktop",
        )
        return name

    @staticmethod
    def get_result_object_name_from_studycase(studycase):
        name = powfacpy.PFTranslator._get_language_dependent_name_from_studycase(
            studycase=studycase,
            english_name="All calculations.ElmRes",
            german_name="Alle Berechnungsarten.ElmRes",
        )
        return name

    @staticmethod
    def get_name_with_ending(objects):
        if not type(objects) == list:
            objects = [objects]
        return [x.GetFullName().split("\\")[-1] for x in objects]



"""Wrapper for the native API of PowerFactory.
"""

from .deprecated.folders_interface import *
from .deprecated.active_project_interface import *
from .deprecated.dyn_sim_interface import *
from .deprecated.plot_interface import *
from .deprecated.case_studies import *
from .deprecated.network_interface import *
from .deprecated.results_interface import *
from .deprecated.script_interface import *
from .deprecated.database_interface import *
from .deprecated.model_exchange_interfaces import *

from .base.active_project import ActiveProject, ActiveProjectCached



import powfacpy
import sys
from warnings import warn

sys.path.insert(0, r".\src")


class PFNetworkInterface(powfacpy.PFActiveProject):

    def __init__(self, app):
        warn(
            f"{self.__class__.__name__} will be deprecated. Please use the class 'Networks' from 'applications/networks' instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(app)

    def get_vacant_cubicle_of_terminal(self, terminal, new_cubicle_name=None):
        """Gets the first vacant cubicle found in a terminal (i.e. nothing is connected
        to this cubicle).
        If there is no vacant cubicle, a new cubicle is created.

        Arguments:
          terminal: ElmTerm
          new_cubicle_name: Name  that is set for the found or created cubicle
        """
        terminal = self._handle_single_pf_object_or_path_input(terminal)
        cubicles = self.get_cubicles_of_terminal(terminal)
        for cubicle in cubicles:
            if cubicle.obj_id == None:
                if new_cubicle_name:
                    cubicle.loc_name = new_cubicle_name
                return cubicle
        if not new_cubicle_name:
            new_cubicle_name = "Cub_" + str(len(cubicles) + 1)
        return self.create_in_folder(new_cubicle_name + ".StaCubic", terminal)

    def get_cubicles_of_terminal(self, terminal, only_calc_relevant=False):
        if not only_calc_relevant:
            return self.get_obj(
                "*",
                parent_folder=terminal,
                condition=lambda x: x.GetClassName() == "StaCubic",
            )
        else:
            return terminal.GetCalcRelevantCubicles()

    def get_elements_connected_to_terminal(
        self, terminal, only_calc_relevant: bool = False
    ):
        cubicles = self.get_cubicles_of_terminal(
            terminal, only_calc_relevant=only_calc_relevant
        )
        elements = []
        for cub in cubicles:
            if cub.obj_id is not None:
                elements.append(cub.obj_id)
        return elements

    def get_connected_terminal(self, element):
        return element.bus1.cterm

    def copy_grid(
        self,
        grid_or_path,
        target_folder,
        new_name,
        parent_folder=None,
        error_if_non_existent=True,
    ):
        """Copying a grid is not trivial in PF because the graphical network objects need to be copied and assigned manually as this is not done automatically."""
        grid_to_be_copied = self._handle_single_pf_object_or_path_input(grid_or_path)
        new_grid = self.copy_single_obj(
            grid_to_be_copied,
            target_folder,
            overwrite=True,
            new_name=new_name,
            parent_folder=parent_folder,
            error_if_non_existent=error_if_non_existent,
        )
        new_network_diagram = self.copy_single_obj(
            grid_to_be_copied.pDiagram,
            self.app.GetProjectFolder("dia"),
            new_name=new_name,
            overwrite=True,
        )
        graphical_net_objects = self.get_obj(
            "*.IntGrf", parent_folder=new_network_diagram, include_subfolders=True
        )
        for graphical_net_obj in graphical_net_objects:
            element = graphical_net_obj.pDataObj
            path_in_grid = self.get_path_between_objects(grid_to_be_copied, element)
            graphical_net_obj.pDataObj = self.get_unique_obj(
                path_in_grid, parent_folder=new_grid
            )
        new_network_diagram.pDataFolder = new_grid
        new_grid.pDiagram = new_network_diagram
        return new_grid

    def get_parent_grid(self, obj_or_path):
        obj_or_path = self._handle_single_pf_object_or_path_input(obj_or_path)
        return self.get_upstream_obj(obj_or_path, lambda x: x.GetClassName == "IntNet")



"""Module to cache attributes and methods of PF objects to improve performance. 
"""

from powfacpy.pf_classes.protocols import PFGeneral


def cache_attr(obj: PFGeneral, attr: str):
    obj.__dict__[attr] = obj.__getattr__(attr)


def cache_attrs(obj: PFGeneral, attrs: list[str]):
    for attr in attrs:
        obj.__dict__[attr] = obj.__getattr__(attr)


def cache_method(obj: PFGeneral, method: str):
    return_val = getattr(obj, method)()
    obj.__dict__[method] = lambda: return_val


def cache_methods(obj: PFGeneral, methods: list[str]):
    for method in methods:
        cache_method(obj, method)



from __future__ import annotations
import sys
from typing import Iterable, Union
from fnmatch import fnmatch
from collections.abc import Generator, Callable
from math import inf

from icecream import ic

sys.path.insert(0, r".\src")
import powfacpy
from powfacpy.pf_class_protocols import *


class PFSet:

    def __init__(self, set: SetSelect) -> None:
        self.set: SetSelect = set

    def get_all_elements(self):
        return self.set.All()


class GroupOfObjects:

    def __init__(self, objects: Iterable[PFGeneral] | Callable) -> None:
        self.objs = objects

    @property
    def objs(self) -> Iterable[PFGeneral]:
        if self._objs:
            return self._objs
        else:
            return self._call_objs()

    @objs.setter
    def objs(self, objects: Iterable[PFGeneral] | Callable):
        if isinstance(objects, Callable):
            self._call_objs: Callable = objects
            self._objs: Iterable[PFGeneral] = []
        else:
            self._objs: Iterable[PFGeneral] = objects

    def __iter__(self):
        for o in self.objs:
            yield o

    def __getitem__(self, obj) -> Union[list[PFGeneral], PFGeneral]:
        return self.get_by_loc_name_with_class(obj)

    def cache(self):
        self._objs = self._call_objs()

    def free_cache(self):
        self._objs = []

    def freeze(self):
        self._objs = set(self._objs)

    def append(self, obj: PFGeneral):
        self.objs.append(obj)

    def get_by_loc_name(self, obj: str) -> filter:
        return filter(lambda x: fnmatch(x.loc_name, obj), self.objs)

    def get_by_loc_name_with_class(self, obj: str) -> filter:
        return filter(
            lambda x: fnmatch(x.loc_name + "." + x.GetClassName(), obj), self.objs
        )

    def get_by_path_in_project(self, obj: str, pf_app) -> filter:
        return filter(
            lambda x: fnmatch(
                powfacpy.PFStringManipulation.format_full_path(str(x), pf_app), obj
            ),
            self.objs,
        )

    def filter(self, function: Callable):
        return filter(function, self.objs)


class Subsystem:

    def __init__(
        self, elms: list[PFGeneral] = [], name: str = "", parent: Subsystem = None
    ) -> None:
        self.name: str = name
        self.elms: list[PFGeneral] = elms
        self.parent: Subsystem = parent
        self.children: list[Subsystem] = []
        self.categories: list[str] = []

    def add_subsystem(self, subsystem: Subsystem):
        self.children.append(subsystem)
        self.elms = set(self.elms) - set(subsystem.elms)

    def get_subsystems(
        self,
        include: Callable[[Subsystem], bool] = lambda x: True,
        stop_at: Callable[[Subsystem], bool] = lambda x: False,
        depth=inf,
        start_depth=0,
    ) -> Generator[Subsystem]:
        if depth > 0:
            for ss in self.children:
                if start_depth < 1:
                    if not stop_at(ss):
                        if include(ss):
                            yield ss
                        for subsubsystem in ss.get_subsystems(
                            include,
                            stop_at,
                            depth=depth - 1,
                            start_depth=start_depth - 1,
                        ):
                            yield subsubsystem

    def get_subsytems_by_names(
        self, names, stop_at_names, depth: int = inf, start_depth=0
    ) -> Generator[Subsystem]:
        return self.get_subsystems(
            lambda x: x.name in names, lambda x: x.name in stop_at_names
        )

    def get_subsystems_by_categories(
        self,
        categories: list[str],
        stop_at_categories: list[str],
        depth: int = inf,
        start_depth=0,
    ) -> Generator[Subsystem]:
        categories_fun = lambda x: any([x in categories for x in self.categories])
        stop_at_categories_fun = lambda x: any(
            [x in stop_at_categories for x in self.categories]
        )
        return self.get_subsystems(
            categories_fun, stop_at_categories_fun, depth, start_depth
        )

    def get_subsystems_by_all_categories(
        self,
        categories: list[str],
        stop_at_categories: list[str],
        depth: int = inf,
        start_depth=0,
    ) -> Generator[Subsystem]:

        categories_fun = lambda x: all([x in categories for x in self.categories])
        stop_at_categories_fun = lambda x: any(
            [x in stop_at_categories for x in self.categories]
        )
        return self.get_subsystems(
            categories_fun, stop_at_categories_fun, depth, start_depth
        )

    def get_all_subsystems(self) -> Generator[Subsystem]:
        for ss in self.children:
            yield ss
            for subsubsystem in ss.get_all_subsystems():
                yield subsubsystem

    def get_all_elms(self) -> Generator[PFGeneral]:
        for elm in self.elms:
            yield elm
        for ss in self.get_all_subsystems():
            for elm in ss.get_all_elms():
                yield elm

    def is_in_elms_of(
        self, subsystem: Subsystem, depth: int = inf, start_depth: int = 0
    ):

        if start_depth < 1 and all([x in subsystem.elms for x in self.elms]):
            return subsystem
        elif depth > 0:
            for ss in subsystem.children:
                found_ss = self.is_in_elms_of(
                    ss, depth=depth - 1, start_depth=start_depth - 1
                )
                if found_ss:
                    return found_ss
        return None

    def get_elements_incl_children(self) -> list:
        elements = self.internal_elements
        for subsys in self.children:
            elements += subsys.get_elements_incl_children()
        return elements


class PFZoneInterface(powfacpy.PFActiveProject):

    def __init__(self, pf_app: powfacpy.PFApp, elms: list[PFGeneral] = []) -> None:
        super().__init__(pf_app)
        self.zone_obj: ElmZone

    # def create_zone(self, elms) -> None:



import pytest
import sys

sys.path.insert(0, r".\src")
import powfacpy
import importlib

importlib.reload(powfacpy)

from test_active_project_interface import pfp

sys.path.insert(0, r".\tests")


@pytest.fixture
def pfni(pf_app):
    return powfacpy.PFNetworkInterface(pf_app)


def test_get_vacant_cubicle_of_terminal(pfni, activate_powfacpy_test_project):
    study_case = pfni.get_single_obj(r"Study Cases\test_network_interface\Study Case")
    study_case.Activate()
    pfni.get_vacant_cubicle_of_terminal(
        r"Network Model\Network Data\test_plot_interface\Grid 1\Terminal HV 2"
    )


if __name__ == "__main__":
    pytest.main(([r"tests\test_network_interface.py"]))



from abc import ABC, abstractmethod
from typing import Callable

from powfacpy.base.active_project import ActiveProjectCached
from powfacpy.pf_classes.protocols import PFGeneral, ElmRes, StaPll


class GroupingBase(ABC):
    """Base class for groupings (ElmZone, ElmArea, ElmBoundary). Provides a common interface (e.g. to get internal elements)."""

    __slots__ = ()

    def add_results_variable_for_elms(
        self,
        condition_for_elms: Callable,
        result_variables: str | list[str],
        results_obj: ElmRes | None = None,
    ) -> None:
        act_prj = ActiveProjectCached()
        if not results_obj:
            results_obj = act_prj.get_from_study_case("ElmRes")
        else:
            results_obj = act_prj._handle_single_pf_object_or_path_input(results_obj)
        elms = self.get_internal_elms(condition_for_elms)
        for elm in elms:
            act_prj.add_results_variable(elm, result_variables, results_obj)

    def get_internal_elms(self, condition: Callable | None = None) -> list[PFGeneral]:
        elms = self.get_all_internal_elms()
        if condition:
            act_prj = ActiveProjectCached()
            elms = act_prj.get_by_condition(elms, condition)
        return elms

    def get_internal_elms_of_class(self, class_name: str) -> list[PFGeneral]:
        return self.get_internal_elms(lambda x: x.GetClassName() == class_name)

    def get_phase_locked_loops(self):
        act_prj = ActiveProjectCached()
        all_plls: list[StaPll] = act_prj.get_calc_relevant_obj("*.StaPll")
        internal_terminals = set(self.get_internal_elms_of_class("ElmTerm"))
        return [pll for pll in all_plls if pll.pbusbar in internal_terminals]

    @abstractmethod
    def get_all_internal_elms(self) -> list[PFGeneral]:
        pass



from __future__ import annotations

from powfacpy.pf_classes.protocols import ElmZone, PFGeneral
from powfacpy.pf_classes.elm.elm_base import ElmBase
from powfacpy.pf_classes.elm.grouping_base import GroupingBase
from powfacpy.result_variables import ResVar

RMS_BAL = ResVar.RMS_Bal
LF_BAL = ResVar.LF_Bal


class ZoneStatic(ElmBase, GroupingBase):

    __slots__ = ()

    def __init__(self, obj: ElmZone) -> None:
        super().__init__(obj)
        self._obj: ElmZone

    def __new__(cls, *args, **kwargs) -> ElmZone | Zone:
        """Implemented only to add type hints for the instance.

        Returns:
            ElmZone | Zone: New instance
        """
        instance = super().__new__(cls)
        return instance

    def get_all_internal_elms(
        self,
    ) -> list[PFGeneral]:
        return self._obj.GetAll()

    def get_internal_elms_of_class(self, class_name: str) -> list[PFGeneral]:
        return self._obj.GetObjs(class_name)

    @staticmethod
    def get_P_exchange_res_var_lf_bal() -> str:
        return LF_BAL.ElmZone.c_InterP.value

    @staticmethod
    def get_Q_exchange_res_var_lf_bal() -> str:
        return LF_BAL.ElmZone.c_InterQ.value

    @staticmethod
    def get_P_exchange_res_var_rms_bal() -> str:
        return RMS_BAL.ElmZone.c_Pinter.value

    @staticmethod
    def get_Q_exchange_res_var_rms_bal() -> str:
        return RMS_BAL.ElmZone.c_Qinter.value


class Zone(ZoneStatic):
    pass



import pytest
import sys

sys.path.insert(0, r".\src")
import powfacpy
import importlib

importlib.reload(powfacpy)
from matplotlib import pyplot
from os import getcwd
import os

from test_active_project_interface import pfp


@pytest.fixture
def pfplot(pf_app):
    # Return PFPlotInterface instance and show app
    pfplot = powfacpy.PFPlotInterface(pf_app)

    return pfplot


@pytest.fixture(autouse=True)
def run_around_tests(pfplot):
    # Code run before each test:
    # PF app needs to be shown for some reason, because otherwise
    # PF can error when plotting.
    pfplot.app.Show()
    # Run test
    yield  # gives control back to the 'context manager', i. e. the function that called this fixture.
    # Code that will run after each test:
    # Make sure that app is hidden for other modules than plot interface
    pfplot.app.Hide()


def test_plot(pfplot, activate_powfacpy_test_project):
    pfsim = powfacpy.PFDynSimInterface(pfplot.app)
    pfsim.activate_study_case(r"Study Cases\test_plot_interface\Study Case 1")
    pfsim.initialize_and_run_sim()

    with pytest.raises(powfacpy.exceptions.PFAttributeNotSetError):
        pfplot.plot(
            r"Network Model\Network Data\test_plot_interface\Grid 1\AC Voltage Source",
            ["s:u0", "m:Qsum:bus1"],
        )

    pfplot.clear_plot_pages()
    pfplot.set_active_plot("test_plot 1", "test_plot_interface 1")
    pfplot.plot(
        r"Network Model\Network Data\test_plot_interface\Grid 1\AC Voltage Source",
        ["s:u0", "m:Qsum:bus1"],
    )

    pfplot.set_active_plot("test_plot 2", "test_plot_interface 1")
    pfplot.plot(
        r"Network Model\Network Data\test_plot_interface\Grid 1\AC Voltage Source",
        "s:u0",
        linestyle=2,
        linewidth=100,
        color=3,
        label="u0 test",
        result_obj=r"Study Cases\test_plot_interface\Study Case 1\All calculations",
    )
    pfplot.set_x_axis_attributes(axisMode=2)
    pfplot.set_y_axis_attributes(scaleType=1)


def test_pyplot_from_csv(pfplot, activate_powfacpy_test_project):
    export_dir = os.getcwd() + r"\tests\tests_output"
    file_name = "results"
    pfsim = powfacpy.PFDynSimInterface(pfplot.app)
    pfsim.activate_study_case(r"Study Cases\test_plot_interface\Study Case 1")
    pfsim.initialize_and_run_sim()
    pfri = powfacpy.PFResultsInterface(pfplot.app)
    pfri.export_to_csv(export_dir, file_name=file_name)

    pyplot.figure()
    powfacpy.PFPlotInterface.plot_from_csv(
        export_dir + "\\" + file_name + ".csv",
        [
            r"test_plot_interface\Grid 1\AC Voltage Source\s:u0",
            r"test_plot_interface\Grid 1\AC Voltage Source\m:Qsum:bus1",
        ],
    )
    pyplot.xlabel("t [s]")


def test_copy_graphics_board_content(pfplot, activate_powfacpy_test_project):
    source_study_case = r"Study Cases\test_plot_interface\Study Case 1"
    target_study_cases = [
        r"Study Cases\test_plot_interface\Study Case 2",
        r"Study Cases\test_plot_interface\Study Case 3",
    ]
    pfplot.copy_graphics_board_content(
        source_study_case,
        target_study_cases,
        obj_to_copy="*.GrpPage",
        clear_target_graphics_board=True,
    )

    pfplot.copy_graphics_board_content(
        source_study_case,
        r"Study Cases\test_plot_interface\Study Case 2",
        obj_to_copy="*.GrpPage",
        clear_target_graphics_board=True,
    )


def test_copy_graphics_board_content_to_all_study_cases(
    pfplot, activate_powfacpy_test_project
):
    source_study_case = r"Study Cases\test_plot_interface\Study Case 1"
    pfplot.copy_graphics_board_content_to_all_study_cases(
        source_study_case, target_parent_folder=r"Study Cases\test_plot_interface"
    )


def test_plot_from_comtrade(pfplot, activate_powfacpy_test_project):
    path_of_cfg = os.path.dirname(__file__) + r"\tests_input\test_comtrade.cfg"
    pfplot.set_active_plot("test_plot 1", "test_plot_interface 1")
    pfplot.plot_from_comtrade(
        path_of_cfg,
        "AC Voltage Source:m:u:bus1:A",
        linestyle=2,
        linewidth=100,
        color=3,
        label="comtrade test",
    )


def test_activate_plot(pfplot, activate_powfacpy_test_project):
    with pytest.raises(powfacpy.exceptions.PFAttributeNotSetError):
        pfplot.set_active_plot("test_plot 1")


def test_clear_curves_by_index_from_active_plot(pfplot, activate_powfacpy_test_project):
    pfplot.get_unique_obj(
        r"Study Cases\test_plot_interface\Study Case 1", include_subfolders=False
    ).Activate()
    pfplot.set_active_plot("test clear curves", "test clear curves")

    # Plot and simulate
    pfplot.clear_curves()
    pfplot.plot(
        r"Network Model\Network Data\test_plot_interface\Grid 1\AC Voltage Source",
        "m:Psum:bus1",
    )
    pfplot.plot(
        r"Network Model\Network Data\test_plot_interface\Grid 1\AC Voltage Source",
        "m:Qsum:bus1",
    )
    pfplot.plot(
        r"Network Model\Network Data\test_plot_interface\Grid 1\AC Voltage Source",
        "s:u0",
    )
    pfdi = powfacpy.PFDynSimInterface(pfplot.app)
    pfdi.initialize_and_run_sim()
    # Clear the last curve
    pfplot.clear_curves_by_index_from_active_plot(slice(-1, 1, -1))
    attributes = pfplot.get_curve_table_attributes()
    assert len(attributes["curveTableResultFile"]) == 2
    assert attributes["curveTableVariable"][0] == "m:Psum:bus1"

    # Plot again
    pfplot.clear_curves()
    pfplot.plot(
        r"Network Model\Network Data\test_plot_interface\Grid 1\AC Voltage Source",
        "m:Psum:bus1",
    )
    pfplot.plot(
        r"Network Model\Network Data\test_plot_interface\Grid 1\AC Voltage Source",
        "m:Qsum:bus1",
    )
    pfplot.plot(
        r"Network Model\Network Data\test_plot_interface\Grid 1\AC Voltage Source",
        "s:u0",
    )
    # Clear the first curve
    pfplot.clear_curves_by_index_from_active_plot(0)
    attributes = pfplot.get_curve_table_attributes()
    assert len(attributes["curveTableResultFile"]) == 2
    assert attributes["curveTableVariable"][0] == "m:Qsum:bus1"


if __name__ == "__main__":
    pytest.main(([r"tests\test_plot_interface.py"]))
    # pytest.main(([r"tests\test_plot_interface.py::test_plot"]))



from __future__ import annotations

from powfacpy.base.active_project import ActiveProjectCached
from powfacpy.pf_classes.protocols import ElmSym, TypSym

from powfacpy.pf_classes.elm.elm_base import ElmBase, SinglePortBase
from powfacpy.result_variables import ResVar


class SynchronousMachine(ElmBase, SinglePortBase):

    __slots__ = ()

    def __init__(self, obj: ElmSym) -> None:
        super().__init__(obj)
        self._obj: ElmSym

    def __new__(cls, *args, **kwargs) -> ElmSym | SynchronousMachine:
        """Implemented only to add type hints for the created instance.

        Returns:
            ElmSym | SynchronousMachine: New instance
        """
        instance = super().__new__(cls)
        return instance

    @property
    def ratedS(self) -> float:
        "Apparent power [MVA]. Parallel machines are considered."
        obj = self._obj
        typ: TypSym = obj.typ_id
        return typ.sgn * obj.ngnum

    @property
    def H_in_pu_based_on_Snom(self) -> float:
        "Inertia constant [s]"
        return self._obj.typ_id.h

    @property
    def J(self) -> float:
        "Moment of Inertia [kgm^2]. Parallel machines are considered."
        obj = self._obj
        return obj.typ_id.J * obj.ngnum

    def get_averaged_internal_reactance(self) -> float:
        """If the machine is not symmetrical, the average of the d-and q-axis internal reactances is used. For example, for a 6th order model, one has xG =0.5(x''d + x''q ).

        Returns:
            float: internal reactance [pu]
        """
        typ: TypSym = self._obj.typ_id
        return 0.5 * (typ.xdss + typ.xqss)

    def get_averaged_internal_susceptance(self) -> float:
        return 1 / self.get_averaged_internal_reactance()

    @staticmethod
    def get_cgmes_mapping():
        return {"inertia": "h"}



"""Classes for interaction with the PF database. Based on folders from where downstream objects can accessed (see class PFFolder). The class PFActiveProject inherits from PFFolder and adds some functionality for the active project.

Also contains the class PFStringManipulation for string manipulation.

The acronym 'PF' is used for 'PowerFactory'.
"""

from __future__ import annotations
from collections.abc import Iterable
from typing import Union, Callable, Generator
from os import path as os_path
from functools import partial
from warnings import warn

from icecream import ic

import powfacpy
import powfacpy.exceptions
from powfacpy.pf_class_protocols import (
    PFApp,
    PFGeneral,
    IntFolder,
    IntPrj,
    SetPrj,
    SetFilt,
)
from powfacpy.base.string_manipulation import PFStringManipulation


class PFFolder:
    """Class for interacting with folders in the PowerFactory database (any folder in the hierarchy).

    The class is very large and has methods to (you can search for sections):
        - Get objects (search for '# Get')
        - Create objects (search for '# Create')
        - Copy objects (search for '# Copy')
        - Delete objects (search for '# Delete')
        - Get/set attributes (search for '# Attributes')
        - Utils to handle user input etc. (search for '# Handle')
        - Utils to handle path strings (search for '# Path')
        - Other functionality (search for '# Other')

    The class is large to have a convenient interface with a lot of functionality to interact with the PF database (at the cost of the clarity of the code unfortunately)
    """

    def __init__(self, folder: Union[PFGeneral, str], pf_app: PFApp):
        warn(
            f"{self.__class__.__name__} will be deprecated. Please use the class 'Folder' from 'base/folder' instead",
            DeprecationWarning,
            stacklevel=2,
        )
        if pf_app:
            self.app: PFApp = pf_app
        else:
            raise TypeError(
                "The input app is of type 'NoneType'. Maybe the PowerFactory app was not loaded correctly."
            )
        self._folder = self.app.GetActiveProject()
        if not self._folder:
            self._folder = self.app.GetCurrentUser()
        self._folder: PFGeneral = self._handle_single_pf_object_or_path_input(folder)

    def __getitem__(self, obj: str) -> PFGeneral:
        """Allows to call 'get_obj' (without optional arguments, i.e. subfolders are included) using brakets.

        Example: terminals = folder_instance['*.ElmTerm']
            where folder_instance is an instance of this class.

        Args:
            obj (str): Object(s) to search for

        Returns:
            PFGeneral: Pointers to object(s) from the PF database
        """
        return self.get_obj(obj, include_subfolders=True)

    def __iter__(self) -> Generator:
        for obj in self.get_obj(
            "*", include_subfolders=False, error_if_non_existent=False
        ):
            yield obj

    def __str__(self) -> str:
        return str(self._folder)

    def __len__(self) -> int:
        return len(self._folder.GetContents("*"))

    ##################
    # Reimplemenatation of methods of PF folder/container objects (to mimic the behavior of powerfactory.DataObject).
    ##################

    def GetContents(self, name: str, recursive: int = 0) -> list[PFGeneral]:
        return self._folder.GetContents(name, recursive)

    def GetChildren(
        self, hiddenMode: int, filter: str, subfolders: int
    ) -> list[PFGeneral]:
        return self._folder.GetChildren(hiddenMode, filter, subfolders)

    def GetAttribute(self, attr: str) -> Union[int | float | str | PFGeneral | list]:
        return self._folder.GetAttribute(attr)

    def SetAttribute(
        self, attr: str, value: Union[int | float | str | PFGeneral | list]
    ) -> None:
        return self._folder.SetAttribute(attr, value)

    def GetParent(self) -> PFGeneral:
        return self._folder.GetParent()

    def CreateObject(
        self, className: str, objectNameParts: Union[int, str] = ""
    ) -> PFGeneral:
        return self._folder.CreateObject(className, objectNameParts)

    def Delete(self) -> int:
        return self._folder.Delete()

    def AddCopy(
        self,
        objectToCopy: list[PFGeneral] | PFGeneral,
        partOfName: Union[str, int] = "",
    ) -> PFGeneral:
        return self._folder.AddCopy(objectToCopy, partOfName)

    def GetFullName(self) -> str:
        return self._folder.GetFullName()

    ##################
    # Get
    ##################
    def get_obj(
        self,
        path: str,
        condition: Callable = None,
        parent_folder: Union[PFGeneral, PFFolder, str] = None,
        error_if_non_existent: bool = True,
        include_subfolders: bool = False,
    ) -> list[PFGeneral]:
        """Get the PowerFactory object(s) under 'path'. By default, 'path' is relative to 'self._folder' (or the active project folder for instances of PFActiveProject).

        See also 'get_unique_obj'.

        Args:
            path (str): path to object(s), can contain wildcards ("*") after the last "\"

            condition (Callable, optional): filter for attributes etc.; example: "condition = lambda x : getattr(x,"uknom")==110; defaults to None

            parent_folder (Union[PFGeneral, PFFolder, str], optional): specify a parent folder/container (or its path) from where the search is started ('path' is then relativ to 'parent_folder'); defaults to _folder/the active project folder for instances of PFActiveProject

            error_if_non_existent (bool, optional): raise exception if no objects are found; defaults to True

            include_subfolders (bool, optional): include subfolders in the search; defaults to True

        Raises:
            TypeError: If 'path' is not a string

        Returns:
            list[PFGeneral]: List with PF objects

        Examples:
          pfbi.get_obj("Network Model\\Network Data\\Grid\\Terminal 1")

        The path can also start with "\\" and contain wildcards after last "\\":
          get_obj("\\Network Model\\Network Data\\Grid\\*.ElmTerm")

        With condition:
          pfbi.get_obj("Network Model\\Network Data\\Grid\\*.ElmTerm"",
            condition = lambda x : getattr(x,"uknom")==110)

        Note that you can also use "r" at the beginning of the string
        argument to use single "\".
        """
        parent_folder = (
            self._folder
            if not parent_folder
            else self._handle_single_pf_object_or_path_input(
                parent_folder, include_subfolders=False
            )
        )

        if not include_subfolders:
            try:
                obj = parent_folder.GetContents(path)
            except RuntimeError:
                raise TypeError("Path must be of type string.")
        else:
            obj = self._handle_inclusion_of_subfolders(
                path, parent_folder, error_if_non_existent
            )
        if not obj:
            return self._handle_non_existing_obj(
                path, parent_folder, error_if_non_existent
            )
        if condition:
            obj_with_condition = self.get_by_condition(obj, condition)
            if obj_with_condition:
                return obj_with_condition
            else:
                return self._handle_condition_of_obj_not_met(
                    path, parent_folder, error_if_non_existent
                )
        else:
            return obj

    def get_obj_including_subfolders(
        self,
        path: str,
        condition: Callable = None,
        parent_folder: Union[PFGeneral, PFFolder, str] = None,
        error_if_non_existent: bool = True,
        include_subfolders: bool = False,
    ) -> list[PFGeneral]:
        """Call 'get_obj' including subfolders."""
        return self.get_obj(
            path,
            condition=condition,
            parent_folder=parent_folder,
            error_if_non_existent=error_if_non_existent,
            include_subfolders=True,
        )

    def get_obj_partial(self, *args, **kwargs) -> Callable:
        """Partial evaluation of 'get_obj'.

        Returns a callable to get_obj.

        Example: a = get_obj_lazy("path\\to\\obj", error_if_non_existent=False)
                 a() # This calls 'get_obj' with given arguments

        For further information arguments etc. see 'get_obj'.
        """
        return partial(self.get_obj, *args, **kwargs)

    def get_unique_obj(
        self,
        path: str,
        parent_folder: Union[PFGeneral, PFFolder, str] = None,
        error_if_non_existent: bool = True,
        include_subfolders: bool = False,
    ) -> PFGeneral:
        """Get unique PowerFactory object under 'path'.

        Use this method if you want to access one single unique object.
        This method is an alterntive to 'get_obj' and returns the unique object instead of a list (that needs to be accessed with '[0]'). It also checks whether the found object is unique (only one object is found).

        This method is equal to get_single_obj and was added because the method
        name fits better to what the method does (i.e. to get a unique object and to throw an error if the object is not unique).

        Args:
            path (str): path to object(s); can contain wildcards ("*") after the last "\"

            condition (Callable, optional): filter for attributes etc.; example: "condition = lambda x : getattr(x,"uknom")==110; defaults to None

            parent_folder (Union[PFGeneral, PFFolder, str], optional):
            specify a parent folder/container (or its path) from where the search is started ('path' is then relativ to 'parent_folder'); defaults to _folder/the active project folder for instances of PFActiveProject

            error_if_non_existent (bool, optional): raise exception if no objects are found; defaults to True

            include_subfolders (bool, optional): include subfolders in the search; defaults to False

        Raises:
            TypeError: If 'path' is not a string or if several objects were found.

        Returns:
            PFGeneral: PF object
        """
        return self.get_single_obj(
            path,
            parent_folder=parent_folder,
            error_if_non_existent=error_if_non_existent,
            include_subfolders=include_subfolders,
        )

    def get_single_obj(
        self,
        path: str,
        parent_folder: Union[PFGeneral, PFFolder, str] = None,
        error_if_non_existent: bool = True,
        include_subfolders: bool = False,
    ) -> PFGeneral:
        """Get unique PowerFactory object under 'path'.

        Use this method if you want to access one single unique object.
        This method is an alterntive to 'get_obj' and returns the unique object instead of a list (that needs to be accessed with '[0]'). It also checks whether the found object is unique (only one object is found).

        Args:
            path (str): path to object(s);
            can contain wildcards ("*") after the last "\"

            condition (Callable, optional):
            filter for attributes etc.; example: "condition = lambda x : getattr(x,"uknom")==110;
            defaults to None

            parent_folder (Union[PFGeneral, PFFolder, str], optional):
            specify a parent folder/container (or its path) from where the search is started ('path' is then relativ to 'parent_folder');
            defaults to _folder/the active project folder for instances of PFActiveProject

            error_if_non_existent (bool, optional):
            raise exception if no objects are found;
            defaults to True

            include_subfolders (bool, optional):
            include subfolders in the search;
            defaults to True

        Raises:
            TypeError:
            If 'path' is not a string.
            If several objects were found.

        Returns:
            PFGeneral: PF object
        """
        obj = self.get_obj(
            path,
            parent_folder=parent_folder,
            error_if_non_existent=error_if_non_existent,
            include_subfolders=include_subfolders,
        )
        if obj:
            if len(obj) < 2:
                return obj[0]
            else:
                if parent_folder:
                    parent_folder_str = (
                        f" in '{self.get_path_of_object(parent_folder)}'"
                    )
                else:
                    parent_folder_str = ""
                raise TypeError(
                    f"The path {path+parent_folder_str} is not a unique object. Did you use wildcards ('*')? This method only returns single unique objects."
                )
        else:
            return None

    def get_multiple_obj_from_similar_sub_directories(
        self, parent_folders: Union[list[PFGeneral], str], sub_path: str
    ) -> list[PFGeneral]:
        """Get multiple objects that are in a similar subdirectory relativ to
        their parent folders

        Args:
            parent_folders (Union[list[PFGeneral], str]): parent folders

            sub_path (str):
                path within the parent folders (string). Must be  unique (don't use placeholders '*')

        Returns:
            list[PFGeneral]: list of PF objects

        Example:
          If you want to get the "All calculation" objects of all the study cases in the study case folder, use
            'self.get_multiple_obj_from_similar_sub_directories(
              'Study Cases\\','All calculations')'
        """
        if isinstance(parent_folders, str):
            parent_folders = self.get_obj(parent_folders)
        objs = []
        for parent in parent_folders:
            parent = self._handle_single_pf_object_or_path_input(parent)
            objs.append(self.get_single_obj(sub_path, parent_folder=parent))
        return objs

    def get_upstream_obj(
        self,
        start_obj_or_path: Union[PFGeneral, str],
        condition: Callable,
        error_if_non_existent: bool = True,
    ) -> PFGeneral:
        """Get upstream object that meets the condition.

        Searches upstream in the PF database hierarchy.

        Args:
            start_obj_or_path (Union[PFGeneral, str], optional): Object/folder (or its path) to start from

            condition (Callable): Condition for the searched object

            error_if_non_existent (bool, optional): If True, an exception is raised if no upstream object is found. Defaults to True.

        Returns:
            PFGeneral: The PF object found
        """
        start_obj_or_path = (
            self._folder
            if not start_obj_or_path
            else self._handle_single_pf_object_or_path_input(start_obj_or_path)
        )
        obj_or_path = start_obj_or_path.GetParent()
        if obj_or_path:
            if condition(obj_or_path):
                return obj_or_path
            else:
                return self.get_upstream_obj(
                    obj_or_path, condition, error_if_non_existent=error_if_non_existent
                )
        else:
            if error_if_non_existent:
                raise Exception(
                    "There is no upstream object that fullfills the condition."
                )
            else:
                return None

    def get_by_condition(
        self, objects: list[PFGeneral], condition: Callable
    ) -> list[PFGeneral]:
        """Get objects (from a list of objects) which satisfy 'condition'.

        Args:
            objects (list[PFGeneral]): list of objects
            condition (Callable): e.g. lambda function

        Raises:
            powfacpy.PFAttributeError: If 'condition' queries attributes which an item in 'objects' does not have.
            TypeError: ToDo

        Returns:
            list[PFGeneral]: PF objects

        Example:
          pfbi.get_by_condition(list_of_objects, lambda x : getattr(x,"uknom")==110)
        """
        objects_true = []
        for obj in objects:
            try:
                # This anonymous function is problematic because it does
                # not always throw an error when it the user provided
                # an anonymous function that does not make sense.
                if condition(obj):
                    objects_true.append(obj)
            except AttributeError as e:
                raise powfacpy.exceptions.PFAttributeError(obj, e, self)
            except TypeError as e:
                object_str = self.get_path_of_object(obj)
                raise TypeError(
                    f"{e}. Maybe an unexpected type is used "
                    f"for attribute of object '{object_str}'."
                )
        return objects_true

    ##################
    # Create
    ##################
    def create_by_path(self, path: str, overwrite: bool = True) -> PFGeneral:
        """Create an object by specifying its path (including the class at the end) and return the object.

        Args:
            path (str): path including class name at th end
            overwrite (bool, optional): Overwrite extisting object. Defaults to True.

        Raises:
            TypeError: If 'path' is not a string

        Returns:
            PFGeneral: The created object

        Example:
          pfbi.create_by_path('Library\\Dynamic Models\\dummy.BlkDef')
        """
        try:
            folder_path, obj_name_incl_class = path.rsplit("\\", 1)
        except AttributeError:
            raise TypeError("The argument 'path' must be of type string.")
        folder = self.get_single_obj(folder_path)
        return self.create_in_folder(obj_name_incl_class, folder, overwrite=overwrite)

    def create_in_folder(
        self,
        obj: str,
        folder: Union[PFGeneral, PFFolder, str] = None,
        overwrite: bool = True,
        use_existing: bool = False,
    ) -> PFGeneral:
        """Create an object inside a folder and return the object.

        Args:
            obj (str): obj including class, e.g. 'model.BlkDef'

            folder (Union[PFGeneral, PFFolder, str], optional): Target folder. Defaults to None (i.e. 'self._folder')

            overwrite (bool, optional): objects with the same name will be overwritten. Defaults to True.

            use_existing (bool, optional):
                Only relevant if 'overwrite' is False.
                If use_existing is False and an object with the same name exists, a new object with "(1)"/"(2)".. in its loc_name is created.
                If use_existing is True and an object with the same name exists, the method just returns the existing object.
                Defaults to False.

        Raises:
            TypeError: If 'obj' is not a string

        Returns:
            PFGeneral: The created object

        Example:
          pfbi.create_in_folder("dummy2.BlkDef", "Library\\Dynamic Models",)
        """
        folder = (
            self._folder
            if not folder
            else self._handle_single_pf_object_or_path_input(folder)
        )
        try:
            obj_name, _, class_name = obj.rpartition(".")
        except AttributeError:
            raise TypeError("The argument 'obj' must be of type string.")
        if overwrite:
            self.delete_obj(
                obj,
                parent_folder=folder,
                include_subfolders=False,
                error_if_non_existent=False,
            )
        elif use_existing:
            existing_obj = self.get_single_obj(
                obj, parent_folder=folder, error_if_non_existent=False
            )
            if existing_obj:
                return existing_obj
        return folder.CreateObject(class_name, obj_name)

    def create_directory(
        self, directory: str, parent_folder: Union[PFGeneral, PFFolder, str] = None
    ) -> IntFolder:
        """Create a directory of folders ('IntFolder') if the directory does not exist yet.

        Args:
            directory (str): path with folders
                Example: 'folder1\\folder2\\folder3'

            parent_folder (Union[PFGeneral, PFFolder, str], optional): Folder to start from. Defaults to None.

        Returns:
            IntFolder: the folder in the lowest subdirectory.
        """
        if not self.path_exists(directory, parent=parent_folder):
            folder = self._folder if not parent_folder else parent_folder
            folder_names = directory.split("\\")
            for folder_name in folder_names:
                if not self.path_exists(folder_name, parent=folder):
                    folder = self.create_in_folder(folder_name + ".IntFolder", folder)
                else:
                    folder = self.get_single_obj(
                        folder_name, parent_folder=folder, include_subfolders=False
                    )
            return folder
        else:
            return self.get_single_obj(
                directory, parent_folder=parent_folder, include_subfolders=False
            )

    def create_filter_obj(
        self,
        name: str,
        folder: str | PFGeneral,
        object_filter: str,
        look_in: PFGeneral | str,
        expression: str,
        include_subfolders: bool = True,
        only_calc_relevant_obj: bool = False,
        overwrite=True,
        use_existing=False,
    ) -> SetFilt:
        """Create filter object (SetFilt)

        Args:
            name (str): Name of filter object.

            folder (PFGeneral | PFFolder | str): target folder where filer is created.

            object_filter (str): Object/class filter (parameter 'objset'). Class names sparated by commas, e.g. '*.ElmTerm, *.ElmLne'.

            look_in (PFGeneral | str): Folder from which search is started (parameter 'pstart').

            expression (str): filter expression (parameter 'expr'), e.g. 'uknom>100.and.uknom<380.and.iUsage=0'

            include_subfolders (bool, optional): Include subfolders in search. Defaults to True.

            only_calc_relevant_obj (bool, optional): Search only for calc. relevant objects. Defaults to True.

            overwrite (bool, optional): Overwrite possible existing filter object. Defaults to True.

            use_existing (bool, optional): See description of 'create_in_folder' in 'PFFolder' class of powfacpy. Defaults to False.

        Returns:
            SetFilt: Filter object
        """

        filter_obj: SetFilt = self.create_in_folder(
            name + ".SetFilt",
            folder=folder,
            overwrite=overwrite,
            use_existing=use_existing,
        )
        look_in = self._handle_single_pf_object_or_path_input(look_in)
        filter_obj.objset = [object_filter]
        filter_obj.expr = expression
        filter_obj.pstart = look_in
        filter_obj.isubfold = include_subfolders
        filter_obj.icalcrel = only_calc_relevant_obj
        return filter_obj

    ##################
    # Copy
    ##################
    def copy_obj(
        self,
        obj_or_path: Union[PFGeneral, str],
        target_folder: Union[PFGeneral, PFFolder, str],
        overwrite: bool = True,
        condition: Callable = None,
        parent_folder: Union[PFGeneral, PFFolder, str] = None,
        error_if_non_existent: bool = True,
        include_subfolders: bool = False,
    ) -> list[PFGeneral]:
        """Copy object(s) to 'target_folder'.

        Uses 'get_obj' to get the objects under 'obj_or_path'. Source and target must be in the active project (otherwise use PasteCopy(), see scripting reference)

        See also 'copy_single_obj'.

        Args:
            obj_or_path (Union[PFGeneral, str]):
                Objects to be copied (get_object is used)

            target_folder (Union[PFGeneral, PFFolder, str]):
                Folder to which objects are copied

            overwrite (bool, optional):
                Overwrite existing objects. Defaults to True.

            condition (Callable, optional):
                Condition used in 'get_obj' for the source objects . Defaults to None.

            parent_folder (Union[PFGeneral, PFFolder, str], optional):
                refers to the source folder and is used in combination with 'obj_or_path' to get the object(s) to be copied. Defaults to  None (i.e. '_folder'/active project is used).

            error_if_non_existent (bool, optional):
                Raise error if no (source) objects are found. Defaults to True.

            include_subfolders (bool, optional):
                Include subfolder in search for source objects. Defaults to False.

        Returns:
            list[PFGeneral]: Copied objects
        """
        obj = self._handle_pf_object_or_path_input(
            obj_or_path,
            condition=condition,
            parent_folder=parent_folder,
            error_if_non_existent=error_if_non_existent,
            include_subfolders=include_subfolders,
        )
        target_folder = self._handle_single_pf_object_or_path_input(target_folder)
        if overwrite:
            for object_to_be_copied in obj:
                self.delete_obj(
                    object_to_be_copied.GetAttribute("loc_name"),
                    parent_folder=target_folder,
                    error_if_non_existent=False,
                )
        # AddCopy() accepts a list of objects, but then it returns the target folder object and not the copied objects. Therefore, it is iterated through the objects.
        copied_obj = []
        for o in obj:
            copied_obj.append(target_folder.AddCopy(o))
        return copied_obj

    def copy_single_obj(
        self,
        obj_or_path: Union[PFGeneral, str],
        target_folder: Union[PFGeneral, PFFolder, str],
        overwrite: bool = True,
        new_name: str = None,
        parent_folder: Union[PFGeneral, PFFolder, str] = None,
        error_if_non_existent: bool = True,
    ) -> PFGeneral:
        """Copy a single PF object to 'target_folder'

         Uses 'get_unique_obj' to get the object under 'obj_or_path'. Source and target must be in the active project (otherwise use PasteCopy(), see scripting reference)

        See also 'copy_obj'.

        Args:
            obj_or_path (Union[PFGeneral, str]):
                Object to be copied (get_unique_object is used)

            target_folder (Union[PFGeneral, PFFolder, str]):
                Folder to which object is copied

            overwrite (bool, optional):
                Overwrite existing objects. Defaults to True.

            new_name (str, optional):
                New name (if different to original). Defaults to None.

            parent_folder (Union[PFGeneral, PFFolder, str], optional):
                refers to the source folder and is used in combination with 'obj_or_path' to get the object(s) to be copied. Defaults to  None (i.e. '_folder'/active project is used).

            error_if_non_existent (bool, optional):
                Raise error if no (source) objects are found. Defaults to True.

            include_subfolders (bool, optional):
                Include subfolder in search for source objects. Defaults to False.

        Returns:
            PFGeneral: Copied object/existing object if overwrite is False
        """
        obj = self._handle_single_pf_object_or_path_input(
            obj_or_path,
            parent_folder=parent_folder,
            error_if_non_existent=error_if_non_existent,
        )
        target_folder = self._handle_single_pf_object_or_path_input(target_folder)
        if overwrite:
            if not new_name:
                self.delete_obj(
                    obj.GetAttribute("loc_name"),
                    parent_folder=target_folder,
                    include_subfolders=False,
                    error_if_non_existent=False,
                )
            else:
                self.delete_obj(
                    f"{new_name}.*",
                    parent_folder=target_folder,
                    include_subfolders=False,
                    error_if_non_existent=False,
                )
        else:
            existing_obj = self.get_single_obj(
                obj.GetAttribute("loc_name"),
                parent_folder=target_folder,
                error_if_non_existent=False,
            )
            if existing_obj:
                return existing_obj
        if new_name:
            return target_folder.AddCopy(obj, new_name)
        else:
            return target_folder.AddCopy(obj)

    ##################
    # Delete
    ##################
    def delete_obj(
        self,
        obj_or_path: Union[PFGeneral, str],
        condition: Callable = None,
        parent_folder: Union[PFGeneral, str] = None,
        error_if_non_existent: bool = True,
        include_subfolders: bool = False,
    ):
        """Delete PF object(s).

        In a first step, the objects are retrieved using the 'get_obj'
        method. In a second step, they are deleted. For further info on the input arguments, see the `get_obj` method. Checks whether objects were really deleted, otherwise tries to deactivate the object and then delete it.

        Args:
            obj_or_path (Union[PFGeneral, str]): objects to be deleted.

            condition (Callable, optional):
                Condition for retrieved object(s). Defaults to None.

            parent_folder (Union[PFGeneral, str], optional):
                Parent folder used in 'get_obj'. Defaults to None.

            error_if_non_existent (bool, optional):
                Throw exception if not objects found. Defaults to True.

            include_subfolders (bool, optional): Search also in subfolders. Defaults to False.

        Raises:
            TypeError: If an object cannot be deleted.
        """
        if obj_or_path:
            obj = self._handle_pf_object_or_path_input(
                obj_or_path,
                condition=condition,
                parent_folder=parent_folder,
                error_if_non_existent=error_if_non_existent,
                include_subfolders=include_subfolders,
            )
            for o in obj:
                o.Delete()
                # 'IsDeleted' seems to be the savest way to check whether an object has been deleted.
                if not o.IsDeleted():
                    active_study_case = self.app.GetActiveStudyCase()
                    if active_study_case:
                        active_study_case.Deactivate()
                        o.Delete()
                        active_study_case.Activate()

                    if not o.IsDeleted():
                        try:
                            o.Deactivate()
                            o.Delete()
                        except AttributeError:  # raised when o cannot be deactivated
                            raise TypeError(f"Object {o} cannot be deleted.")

                        if not o.IsDeleted():
                            raise TypeError(f"Object {o} cannot be deleted.")

    def clear_folder(self, folder: Union[PFGeneral, PFFolder, str] = None):
        """Clear all objects inside folder (including hidden objects).

        Args:
            folder (Union[PFGeneral, PFFolder, str], optional): Folder/ container objects or its path. Defaults to None (i.e. '_folder'/active project).
        """
        folder = (
            self._folder
            if not folder
            else self._handle_single_pf_object_or_path_input(folder)
        )
        self.delete_obj(
            "*",
            parent_folder=folder,
            include_subfolders=False,
            error_if_non_existent=False,
        )

    ##################
    # Attributes
    ##################
    def get_attr(
        self,
        obj: Union[PFGeneral, str],
        attr: str,
        parent_folder: Union[PFGeneral, PFFolder, str] = None,
    ) -> Union[int | float | str | PFGeneral | list]:
        """Get the value of an attribute of an object.

        Args:
            obj (Union[PFGeneral, str]): object or its path
            attr (str): attribute name
            parent_folder (Union[PFGeneral, PFFolder, str], optional):
                - parent folder of object. Defaults to None.

        Raises:
            powfacpy.PFAttributeError: If ojbect does not have the specified attribute

        Returns:
            Union[int|float|str|PFGeneral|list]: Attribute value

        Example:
            get_attr(terminal_1, "systype")
        """
        if isinstance(obj, str):
            obj = self.get_single_obj(obj, parent_folder=parent_folder)
        try:
            if not isinstance(attr, list):
                return obj.GetAttribute(attr)
            else:
                attr_values = dict()
                for attribute in attr:
                    attr_values[attribute] = obj.GetAttribute(attribute)
                return attr_values
        except AttributeError as e:
            raise powfacpy.exceptions.PFAttributeError(obj, e, self)

    def get_attr_by_path(
        self, path_with_attr: str
    ) -> Union[int | float | str | PFGeneral | list]:
        """Get attribute using the path including the atrribute name

        Args:
            path_with_attr (str): path including the atrribute name
            Example: 'user\\project\\path\\to\\object\\m:Psum:bus1'

        Returns:
            Union[int|float|str|PFGeneral|list]: Attribute value
        """
        head_tail = os_path.split(path_with_attr)
        return self.get_attr(head_tail[0], head_tail[1])

    def set_attr(
        self,
        obj: Union[PFGeneral, str],
        params: dict,
        parent_folder: Union[PFGeneral, PFFolder, str] = None,
    ) -> None:
        """Set the attribute(s) of an object.

        Args:
            obj (Union[PFGeneral, str]): PF object
            params (dict): atttributes and their values (e.g. {'parameter1':value1, 'parameter2':value2,..})
            parent_folder (Union[PFGeneral, PFFolder, str], optional): Parent folder object. Defaults to None.

        Raises:
            powfacpy.PFAttributeTypeError: If the type of an attribute value is wrong
            powfacpy.PFAttributeError: If an object does not have th especified attribute
        """
        obj = self._handle_single_pf_object_or_path_input(
            obj, parent_folder=parent_folder
        )
        for attr, value in params.items():
            try:
                obj.SetAttribute(attr, value)
            except TypeError as e:
                raise powfacpy.exceptions.PFAttributeTypeError(obj, attr, e, self)
            except AttributeError as e:
                raise powfacpy.exceptions.PFAttributeError(obj, e, self)

    def set_attr_by_path(
        self, path_with_attr: str, value: Union[int | float | str | PFGeneral | list]
    ):
        """Set attribute using the path including the attribute name.

        Args:
            path_with_attr (str): path to ojbect with attribute name at the end
            value (Union[int | float | str | PFGeneral | list]): attribute value

        Example:
          set_attr_by_path(
                "Library\\Dynamic Models\\Linear_interpolation\\desc",
                ["description"])
          Here 'desc' is the name of the attribute.
        """
        head_tail = os_path.split(path_with_attr)
        self.set_attr(head_tail[0], {head_tail[1]: value})

    ##################
    # Handle
    ##################
    def _handle_inclusion_of_subfolders(
        self,
        path: str,
        parent_folder: Union[PFGeneral, PFFolder, str],
        error_if_non_existent: bool,
    ) -> Union[list, PFGeneral]:
        """Handle the inclusion of subfolders in search for objects.

        If subfolders are included, 'GetChildren' must be used instead of 'GetContents'. 'GetChildren' requires the input to be splitted between path and object name.

        Args:
            path (str): Path to object(s)

            parent_folder (Union[PFGeneral, PFFolder,str]): Parent folder where search is started

            error_if_non_existent (bool): Raise exception if no objects are found

        Raises:
            TypeError: If 'path' is not a string

        Returns:
            Union[list, PFGeneral]: Retrieved objects
        """
        try:
            path_folder_list = path.split("\\")
            head, tail = "\\".join(path_folder_list[:-1]), path_folder_list[-1]
        except AttributeError:
            raise TypeError("Path must be of type string.")
        if head:
            new_parent_folder = parent_folder.GetContents(head)
            if new_parent_folder:
                parent_folder = new_parent_folder[0]
            else:
                return self._handle_non_existing_obj(
                    head, parent_folder, error_if_non_existent
                )
        return parent_folder.GetChildren(1, tail, 1)

    def _handle_non_existing_obj(
        self,
        path: str,
        parent_folder: Union[PFGeneral, PFFolder, str],
        error_if_non_existent: bool,
    ) -> list:
        """Handle the attempted access to a non existent object.

        Args:
            path (str): path to object

            parent_folder (Union[PFGeneral, PFFolder, str]): search is started from here

            error_if_non_existent (bool): raise exception if non existent

        Raises:
            powfacpy.PFPathError: if 'error_if_non_existent' is True

        Returns:
            list: empty list if 'error_if_non_existent' is False
        """
        if not error_if_non_existent:
            return []
        else:
            exists_bool, existing_path, non_existing_child = self.path_exists(
                path, parent_folder, return_info=True
            )
            raise powfacpy.exceptions.PFPathError(non_existing_child, existing_path)

    def _handle_condition_of_obj_not_met(
        self,
        path: str,
        parent_folder: Union[PFGeneral, PFFolder],
        error_if_non_existent: bool,
    ) -> list | None:
        """Handle the attempted access to an object with a specific condition, when such object(s) were found, but none of them meets the condition.

        Args:
            path (str): path of the object(s)

            obj (list[Union[PFGeneral, PFFolder]]): list of objects found using path, but none of them satisfies the condition

            error_if_non_existent (bool): raise exception

        Raises:
            powfacpy.PFNonExistingObjectError: If 'error_if_non_existent' is True

        Returns:
            list | None: empty list
        """
        if not error_if_non_existent:
            return []
        else:
            # head, tail = os_path.split(path)
            raise powfacpy.exceptions.PFNonExistingObjectError(
                parent_folder, path, condition=True
            )

    def _handle_pf_object_or_path_input(
        self,
        obj_or_path: Union[PFGeneral, str],
        condition: bool = None,
        parent_folder: Union[PFGeneral, PFFolder, str] = None,
        error_if_non_existent: bool = True,
        include_subfolders: bool = False,
    ) -> PFGeneral | list[PFGeneral]:
        """Handles the input argument 'obj_or_path' when a method accepts either
          - a path string
          - a PF object
          - an iterable (of PF objects)

        Note: Unfortunately, it cannot be checked whether 'obj' is a PF DataObject, because the module powerfactory is not available.

        See also '_handle_single_pf_object_or_path_input'

        Args:
            obj_or_path (Union[PFGeneral, str]): see get_obj

            condition (bool, optional): see get_obj. Defaults to None.
            parent_folder (Union[PFGeneral, PFFolder, str], optional): see get_obj. Defaults to None.

            error_if_non_existent (bool, optional): see get_obj. Defaults to True.

            include_subfolders (bool, optional): see get_obj. Defaults to False.

        Returns:
            PFGeneral:
                - PF object(s) in a list.
                - If 'obj_or_path' is a path string, it returns the PF object(s) of that path.
                - If 'obj_or_path' is a PF object (or multiple), it does nothing and just returns the object(s).
        """
        if isinstance(obj_or_path, str):
            return self.get_obj(
                obj_or_path,
                condition=condition,
                parent_folder=parent_folder,
                error_if_non_existent=error_if_non_existent,
                include_subfolders=include_subfolders,
            )
        elif not isinstance(obj_or_path, Iterable):
            return [obj_or_path]
        elif not isinstance(obj_or_path[0], str):
            return obj_or_path
        else:
            objs = []
            for obj in obj_or_path:
                objs += self.get_obj(obj)
            return objs

    def _handle_single_pf_object_or_path_input(
        self,
        obj_or_path: Union[PFGeneral, str],
        parent_folder: Union[PFGeneral, PFFolder, str] = None,
        error_if_non_existent: bool = True,
        include_subfolders: bool = False,
    ) -> PFGeneral:
        """Handles the input argument when a method accepts either
          - a path string
          - or a PF object
          - but NOT an iterable of PF objects (difference to    _handle_pf_object_or_path_input)

        Note: Unfortunately, it cannot be checked whether 'obj' is a PF DataObject, because the module powerfactory is not available.

        See also method 'handle_pf_object_path_input'

        Args:
            obj (Union[PFGeneral, str]): see get_single_obj

            parent_folder (Union[PFGeneral, PFFolder, str], optional): get_single_obj. Defaults to None.

            error_if_non_existent (bool, optional): get_single_obj. Defaults to True.

        Raises:
            TypeError: If 'obj_or_path' is an iterable (unexpected), a meaningfull error is raised.

        Returns:
            PFGeneral: If 'obj_or_path' is a path string, it returns the PF object of that path.
            If 'obj_or_path' is a PF object, it does nothing and just returns the object.
        """
        if isinstance(obj_or_path, str):
            return self.get_single_obj(
                obj_or_path,
                parent_folder=parent_folder,
                error_if_non_existent=error_if_non_existent,
                include_subfolders=include_subfolders,
            )
        elif isinstance(obj_or_path, Iterable) and not isinstance(
            obj_or_path, powfacpy.PFFolder
        ):
            elements_count = len(obj_or_path)
            if obj_or_path:
                first_obj_type = type(obj_or_path[0])
                try:
                    first_obj_path = self.get_path_of_object(obj_or_path[0])
                    msg_obj = f"The first element is of type {first_obj_type} and its path is {first_obj_path}."
                except AttributeError:
                    msg_obj = f"The first element is of type {first_obj_type}."
                msg = f"Expected a PowerFactory object or a path string. Instead an iterable of length {elements_count} is given. {msg_obj}"
                raise TypeError(msg)
            else:
                msg = f"Expected a PowerFactory object or a path string. Instead an empty object of type '{type(obj_or_path).__name__}' is given."
                raise TypeError(msg)
        else:  # If all former conditions are False, it is assumed that the
            # input already was a PF object.
            return obj_or_path

    ######################
    # Paths
    ######################
    def path_exists(
        self,
        path: str,
        parent: Union[PFGeneral, PFFolder, str] = None,
        return_info: bool = False,
    ) -> bool:
        """Check if the path exists.

        Args:
            path (str): path

            parent (Union[PFGeneral, PFFolder, str], optional): Parent folder where search is started. Defaults to None (i.e. 'self._folder').

            return_info (bool, optional): information about where the path is
            corrupted is returned. Defaults to False.

        Raises:
            powfacpy.PFPathInputError: If path is invalid.

        Returns:
            bool: True if path exists
        """

        parent = (
            self._folder
            if not parent
            else self._handle_single_pf_object_or_path_input(parent)
        )

        splitted_path = path.split("\\")
        if path[0] == "\\" or not splitted_path:
            raise powfacpy.exceptions.PFPathInputError(path)
        existing_path = ""
        child = parent
        for child_name in splitted_path:
            child = child.GetContents(f"{child_name}")
            if child:
                child = child[0]
                existing_path = f"{existing_path}\\{child.loc_name}"
            else:
                if not return_info:
                    return False
                else:
                    parent_path = parent.GetFullName()
                    parent_path = PFStringManipulation.remove_class_names(parent_path)
                    existing_path = f"{parent_path}{existing_path}"
                    non_existent_child_name = child_name
                    return False, existing_path, non_existent_child_name
        return True

    def get_path_of_object(self, obj: Union[PFGeneral, PFFolder]) -> str:
        """Get path relative to 'self._folder' without class names.

        Args:
            obj (PFGeneral): PF object

        Returns:
            str: path
        """
        obj = self._handle_single_pf_object_or_path_input(obj)
        return PFStringManipulation.remove_class_names(
            self.get_path_of_obj_with_class_names(obj)
        )

    def get_path_of_obj_with_class_names(
        self, obj: Union[PFGeneral, PFFolder, str]
    ) -> str:
        """Get path relative to 'self._folder' including class names.

        Args:
            obj (Union[PFGeneral, PFFolder, str]): PF object

        Returns:
            str: path
        """
        _folder_path_without_closing_tag = (
            PFStringManipulation.remove_closing_html_tag_from_path(str(self._folder))
        )
        obj = self._handle_single_pf_object_or_path_input(obj)
        obj_str = PFStringManipulation.remove_closing_html_tag_from_path(str(obj))
        return PFStringManipulation.truncate_beginning(
            obj_str, _folder_path_without_closing_tag
        )

    def get_full_path_of_object(self, obj: Union[PFGeneral, PFFolder]) -> str:
        """Get full path in PF database without class names.

        Args:
            obj (PFGeneral): PF object

        Returns:
            str: path
        """
        obj = self.get_full_path_of_object_with_class_names(obj)
        return PFStringManipulation.remove_class_names(obj)

    def get_full_path_of_object_with_class_names(
        self, obj: Union[PFGeneral, PFFolder, str]
    ) -> str:
        """Get path in PF database including class names.

        Args:
            obj (PFGeneral): PF object

        Returns:
            str: path
        """
        obj = self._handle_single_pf_object_or_path_input(obj)
        return PFStringManipulation.remove_html_tags_from_path(str(obj))

    def get_path_of_object_in_active_project(
        self, obj: Union[PFGeneral, PFFolder, str]
    ) -> str:
        """Get path relative to active project without class names.

        Args:
            obj (PFGeneral): PF object

        Returns:
            str: path
        """
        obj = self.get_path_of_object_in_active_project_with_class_names(obj)
        return PFStringManipulation.remove_class_names(obj)

    def get_path_of_object_in_active_project_with_class_names(
        self, obj: Union[PFGeneral, PFFolder, str]
    ) -> str:
        """Get path relative to active project including class names.

        Args:
            obj (Union[PFGeneral, PFFolder, str]): PF object or its path

        Returns:
            str: path
        """
        active_project_path = PFStringManipulation.remove_closing_html_tag_from_path(
            str(self.app.GetActiveProject())
        )
        obj = self._handle_single_pf_object_or_path_input(obj)
        obj_str = PFStringManipulation.remove_closing_html_tag_from_path(str(obj))
        return PFStringManipulation.truncate_beginning(obj_str, active_project_path)

    def _replace_special_PF_characters_in_path_string(self, path: str) -> str:
        """Replaces special characters '$(ExtDataDir)','$(WorkspaceDir)','$(InstallationDir)' in path with their actual directories."""
        if "$(ExtDataDir)" in path:
            project_settings = self.get_project_settings()
            ext_data_dir = self.get_attr(project_settings, "extDataDir")
            path = path.replace("$(ExtDataDir)", ext_data_dir)
        path = path.replace("$(WorkspaceDir)", self.app.GetWorkspaceDirectory())
        return path.replace("$(InstallationDir)", self.app.GetInstallationDirectory())

    def get_path_between_objects(
        self,
        obj_high: Union[PFGeneral, PFFolder, str],
        obj_low: Union[PFGeneral, PFFolder, str],
    ) -> str:
        """Get path between two objects in the PF database.

        Args:
            obj_high (Union[PFGeneral, PFFolder, str]): object higher in the hierarchy
            obj_low (Union[PFGeneral, PFFolder, str]): object lower in the hierarchy

        Returns:
            str: path between
        """
        obj_high = self._handle_single_pf_object_or_path_input(obj_high)
        obj_high = self.get_path_of_object(obj_high)
        obj_low = self._handle_single_pf_object_or_path_input(obj_low)
        obj_low = self.get_path_of_object(obj_low)
        path = str(obj_low).split(str(obj_high))[1][1:]
        return path

    @staticmethod
    def get_loc_name_with_class(
        objects: Union[PFGeneral, list[PFGeneral]]
    ) -> Union[str, list[str]]:
        """Get local name (loc_name) including class.

        Args:
            objects (Union[PFGeneral, list[PFGeneral]]): PF objects(s)

        Returns:
            Union[str, list[str]]: local name(s)
        """
        is_list = True
        if not type(objects) == list:
            objects = [
                objects,
            ]
            is_list = False

        loc_name_with_class = []
        for powerfactory_object in objects:
            loc_name_with_class.append(
                powerfactory_object.loc_name + "." + powerfactory_object.GetClassName()
            )

        if not is_list:
            loc_name_with_class = loc_name_with_class[0]
        return loc_name_with_class

    #####################
    # Other
    #####################
    def is_pf_class(self, class_name: str) -> bool:
        """Checks if class_name is a valid PF class (using the class ID)."""
        return self.app.GetClassId(class_name) != 0

    def is_container(self, obj: Union[PFGeneral, PFFolder, str]) -> bool:
        """Checks whether a PF object is a container. It is assumed
        that an object is a container if it has the attribute 'contents'.
        """
        obj = self._handle_single_pf_object_or_path_input(obj)
        return obj.HasAttribute("contents")

    def get_project_settings(self) -> SetPrj:
        """Get project settings object."""
        project_settings_folder = self.get_single_obj("*.SetFold")
        return self.get_single_obj("*.SetPrj", parent_folder=project_settings_folder)

    def get_active_project(self) -> IntPrj:
        """Get currently active project. Throw an error if no project is active."""
        active_project = self.app.GetActiveProject()
        if active_project:
            return active_project
        else:
            raise powfacpy.exceptions.PFNotActiveError("a project")



"""
Interface with pandapower.
Tutorial on internal data structure and matrices of pandapower: https://github.com/e2nIEE/pandapower/blob/develop/tutorials/internal_datastructure.ipynb
"""

from abc import abstractmethod
from typing import Callable

import pandapower as pp
from pandapower.converter.powerfactory.validate import validate_pf_conversion
import pandas as pd
import numpy as np
import copy
from icecream import ic

from powfacpy.applications.application_base import ApplicationBase
from powfacpy.pf_classes.protocols import (
    PFGeneral,
    PFApp,
)


class PandapowerInterface(ApplicationBase):

    def __init__(
        self, pf_app: PFApp | None | bool = False, cached: bool = False
    ) -> None:
        super().__init__(pf_app, cached)

    def pf_project_to_pandapower(
        self, project_path_in_user: str = None, json_path: str = None
    ) -> pp.auxiliary.pandapowerNet:
        """
        Convert PowerFactory network to pandapower format.

        see pandapower docs: https://pandapower.readthedocs.io//en//v2.13.1//converter//powerfactory.html

        The method 'from_pfd' that is used deactivates the currently active PF project, so it is reactivated.

        Arguments:
            project_path_in_user (str): path of the project in PowerFactory database relative to user

            json_path (str): path of exported pandapower json file

        Returns:
            pp.auxiliary.pandapowerNet: pandapower network
        """
        active_project = self.act_prj.app.GetActiveProject()
        if project_path_in_user is None:
            project_path_in_user = self.act_prj.get_path_of_object_in_current_user(
                active_project
            )
        net = pp.converter.powerfactory.from_pfd(
            self.app, prj_name=project_path_in_user, path_dst=json_path
        )
        if active_project:
            active_project.Activate()
        return net

    def get_Ybus_matrix(self, net: pp.pandapowerNet, return_deepcopy=True) -> np.matrix:
        """
        Get admittance matrix (sparse).

        If internal bus-branch model does not exist yet, a power flow must be executed.

        Arguments:
            net: pandapower network
            return_deepcopy: If True, a deepcopy of the matrix stored in 'net' is returned.

        Returns:
            np.matrix: admittance matrix
        """
        if not net._ppc or net._ppc["internal"]["Ybus"].size == 0:
            pp.runpp(net)
        if return_deepcopy:
            return copy.deepcopy(net._ppc["internal"]["Ybus"])
        else:
            return net._ppc["internal"]["Ybus"]

    def get_Ybus_frame(self, net: pp.pandapowerNet) -> pd.DataFrame:
        """Get Dataframe with admittance matrix and row (column) labels according to the 'loc_name' of the terminal objects.

        Args:
            net (pp.pandapowerNet): pandapower dataset (exported from PF)

        Returns:
            pd.DataFrame: admittance matrix
        """
        Yb = self.get_Ybus_matrix(net)
        matrix_index_labels = self.get_admittance_matrix_labels(net)
        return pd.DataFrame(
            Yb.todense(), columns=matrix_index_labels, index=matrix_index_labels
        )

    def get_imaginary_Ybus_matrix(
        self, net: pp.pandapowerNet, return_deepcopy=True
    ) -> np.matrix:
        """Get imaginary part of admittance matrix (sparse).

        Args:
            net (pp.pandapowerNet): pandapower dataset (exported from PF)
            return_deepcopy (bool, optional): _description_. Defaults to True.

        Returns:
            np.matrix: imaginary part of admittance matrix (sparse)
        """
        if return_deepcopy:
            return copy.deepcopy(np.imag(self.get_Ybus_matrix(net)))
        else:
            return np.imag(self.get_Ybus_matrix(net))

    def get_admittance_matrix_labels(self, net: pp.pandapowerNet):
        pd2ppc_lookups = net._pd2ppc_lookups["bus"]
        matrix_index_labels = [None] * net._ppc["internal"]["Ybus"].shape[0]
        for ybus_idx in range(len(matrix_index_labels)):
            bus_idx = np.argmax(net._pd2ppc_lookups["bus"] == ybus_idx)
            matrix_index_labels[ybus_idx] = net.bus.iloc[bus_idx, :]["name"]
        return matrix_index_labels

    def get_Bbus_matrix(self, net: pp.pandapowerNet, return_deepcopy=True) -> np.matrix:
        """
        Get Bbus matrix  (sparse) used for dc power flow.

        If internal bus-branch model for dc power flow does not exist yet, a dc power flow must be executed.

        Arguments:
        - net: pandapower network
        """
        if not net._ppc or not net._ppc["internal"]["Bbus"].size == 0:
            pp.rundcpp(net)
        if return_deepcopy:
            return copy.deepcopy(net._ppc["internal"]["Bbus"])
        else:
            return net._ppc["internal"]["Bbus"]

    def get_jacobian_matrix(
        self, net: pp.pandapowerNet, return_deepcopy=True
    ) -> np.matrix:
        """
        Get load flow jacobian matrix.

        Arguments:
        - net: pandapower network
        """
        if not net._ppc:
            pp.runpp(net)
        if return_deepcopy:
            return copy.deepcopy(net._ppc["internal"]["J"])
        else:
            return net._ppc["internal"]["J"]

    def get_connectivity_matrix(
        self, net: pp.pandapowerNet, boolean: bool = False
    ) -> np.array:
        """Get the connectivity (also called adjacency) matrix.

        Indices of connected nodes are True. The connectivity matrix is the admittance matrix where nonzero entries are True, other entries are False.

        Args:
            net (pp.pandapowerNet): pandapower dataset

        Returns:
            np.array[bool]: Connectivity matrix.
        """
        Ybus = net._ppc["internal"]["Ybus"]
        if not boolean:
            connect_mat = np.full(Ybus.shape, 0)
            connect_mat[np.array(Ybus.todense() != 0)] = 1
        else:
            connect_mat = np.full(Ybus.shape, False)
            connect_mat[np.array(Ybus.todense() != 0)] = True
        return connect_mat

    def get_connectivity_frame(
        self, net: pp.pandapowerNet, boolean: bool = False
    ) -> pd.DataFrame:
        con_mat = self.get_connectivity_matrix(net, boolean=boolean)
        matrix_index_labels = self.get_admittance_matrix_labels(net)
        return pd.DataFrame(
            con_mat, columns=matrix_index_labels, index=matrix_index_labels
        )

    def get_difference_between_pf_and_pandapower_dataset(
        self,
        net: pp.auxiliary.pandapowerNet,
    ) -> pd.DataFrame | None:
        """
        Get difference between pandapower dataset (exported from PF) and the original PF.

        Assumes default settings for units in the PF project (see Settings/Units). Note that only a small subset of classes and parameters is checked (see 'get_pandapower_2_pf_parameter_mapping' and 'get_pandapower_2_pf_class_mapping')

        Args:
            net (pp.auxiliary.pandapowerNet): pandapower dataset (exported from PF)

        Returns:
            pd.DataFrame | None: Dataframe with differences. None if there are no differences
        """
        pandapower_2_pf_class_mapping = self.get_pandapower_2_pf_class_mapping()
        pandapower_2_pf_parameter_mapping = self.get_pandapower_2_pf_parameter_mapping()
        differing_objects = {
            "name": [],
            "pf_class": [],
            "pp_parameter": [],
            "pf_value": [],
            "pp_value": [],
        }
        for pp_class, pf_classes in pandapower_2_pf_class_mapping.items():
            df_pp_class = net[pp_class].set_index("name")
            for pf_class in pf_classes:
                pf_objects: list[PFGeneral] = self.app.GetCalcRelevantObjects(
                    f"*.{pf_class}"
                )
                for pp_param, pf_param in pandapower_2_pf_parameter_mapping[
                    pf_class
                ].items():
                    for pf_obj in pf_objects:
                        if isinstance(pf_param, str):
                            pf_value = getattr(pf_obj, pf_param)
                        else:
                            pf_value = pf_param(pf_obj)
                        pp_value = df_pp_class.loc[pf_obj.loc_name, pp_param]
                        if not (pf_value == pp_value):
                            differing_objects["name"].append(pf_obj.loc_name)
                            differing_objects["pf_class"].append(pf_obj.GetClassName())
                            differing_objects["pp_parameter"].append(pp_param)
                            differing_objects["pf_value"].append(pf_value)
                            differing_objects["pp_value"].append(pp_value)
        if not differing_objects["name"]:
            return None
        else:
            return pd.DataFrame(differing_objects)

    def get_pandapower_2_pf_class_mapping(self) -> dict[str, list[str]]:
        """Get mapping between pandapower classes (components) and PowerFactory classes.

        Returns:
            dict[str, list[str]]: Mapping. Note that the values are lists because one pandapower class can have multiple corresponding pf classes.
        """
        return {
            "bus": ["ElmTerm"],
            "line": ["ElmLne"],
            "trafo": ["ElmTr2", "ElmTr3"],
            "impedance": ["ElmZpu", "ElmSind"],
        }

    def get_pandapower_2_pf_parameter_mapping(
        self,
    ) -> dict[str, dict[str, str | Callable]]:
        """Get parameter mapping between pandapower and PowerFactory.

        Only a small subset of all the parameters is currently implemented.

        Returns:
            dict[str, dict[str, str | Callable]]: For each pf class, a mapping between the pandapower and the pf parameter is provided. The pf parameter can be a string or a callable (input is the pf object; used for example to access a parameter of the type).
        """
        return {
            "ElmTerm": {
                "vn_kv": "uknom",
            },
            "ElmLne": {
                "length_km": "dline",
                "parallel": "nlnum",
                "r_ohm_per_km": lambda x: x.typ_id.rline,
                "x_ohm_per_km": lambda x: x.typ_id.xline,
                "c_nf_per_km": lambda x: x.typ_id.cline
                * 1000  # from micro to nano
                * x.typ_id.frnom  # nominal frequency is adapted to 50 Hz by pandapower
                / 50.0,
                "g_us_per_km": lambda x: x.typ_id.gline,
            },
            "ElmTr2": {
                "sn_mva": lambda x: x.typ_id.strn,
                "parallel": "ntnum",
                "vn_hv_kv": lambda x: x.typ_id.utrn_h,
                "vn_lv_kv": lambda x: x.typ_id.utrn_l,
                "vk_percent": lambda x: x.typ_id.uktr,
                "vkr_percent": lambda x: x.typ_id.uktrr,
            },
            "ElmTr3": {
                "parallel": "nt3nm",
                "sn_hv_mva": lambda x: x.typ_id.strn3_h,
                "sn_mv_mva": lambda x: x.typ_id.strn3_m,
                "sn_lv_mva": lambda x: x.typ_id.strn3_l,
                "vn_hv_kv": lambda x: x.typ_id.utrn3_h,
                "vn_mv_kv": lambda x: x.typ_id.utrn3_m,
                "vn_lv_kv": lambda x: x.typ_id.utrn3_l,
                "vk_hv_percent": lambda x: x.typ_id.uktr3_h,
                "vk_mv_percent": lambda x: x.typ_id.uktr3_m,
                "vk_lv_percent": lambda x: x.typ_id.uktr3_l,
                "vkr_hv_percent": lambda x: x.typ_id.uktrr3_h,
                "vkr_mv_percent": lambda x: x.typ_id.uktrr3_m,
                "vkr_lv_percent": lambda x: x.typ_id.uktrr3_l,
            },
            "ElmZpu": {
                "sn_mva": "Sn",
                "rft_pu": "r_pu",
                "xft_pu": "x_pu",
            },
            "ElmSind": {
                "sn_mva": "Sn",
                "rft_pu": lambda x: (x.rrea * x.Sn) / (x.ucn**2),
                "xft_pu": lambda x: (x.xrea * x.Sn) / (x.ucn**2),
            },
        }



from __future__ import annotations
from warnings import warn
from os import replace
from typing import Union, Callable
from os import path as os_path
from warnings import warn

import powfacpy
import powfacpy.deprecated
from powfacpy.deprecated.folders_interface import PFFolder
from powfacpy.pf_class_protocols import (
    PFApp,
    PFGeneral,
    ElmRes,
    ElmNet,
    IntComtrade,
    IntUser,
    IntCase,
    IntEvt,
    ElmZone,
    ComLdf,
    IntMon,
    IntVersion,
    ComPfdimport,
    IntPrj,
    IntScheme,
)


class PFActiveProject(powfacpy.PFFolder):
    """Interface to the currently active project."""

    def __init__(self, pf_app: PFApp):
        warn(
            f"{self.__class__.__name__} will be deprecated. Please use the class 'ActiveProject' from 'base/active_project' instead",
            DeprecationWarning,
            stacklevel=2,
        )
        if pf_app:
            self.app: PFApp = pf_app
        else:
            raise TypeError(
                "The input app is of type 'NoneType'. Maybe the PowerFactory app was not loaded correctly."
            )

    @property
    def _folder(self):
        return self.get_active_project()

    @property
    def network_model_folder(self):
        return self.app.GetProjectFolder("netmod")

    @property
    def network_data_folder(self):
        return self.app.GetProjectFolder("netdat")

    @property
    def operation_scenarios_folder(self):
        return self.app.GetProjectFolder("scen")

    @property
    def variations_folder(self):
        return self.app.GetProjectFolder("scheme")

    @property
    def study_cases_folder(self):
        return self.app.GetProjectFolder("study")

    @property
    def equipment_type_lib_folder(self):
        return self.app.GetProjectFolder("equip")

    @property
    def library_folder(self):
        return self.app.GetProjectFolder("lib")

    @property
    def scripts_folder(self):
        return self.app.GetProjectFolder("scripts")

    @property
    def templates_folder(self):
        return self.app.GetProjectFolder("templ")

    @property
    def zones_folder(self):
        return self.app.GetDataFolder("ElmZone")

    @property
    def areas_folder(self):
        return self.app.GetDataFolder("ElmArea")

    @property
    def boundaries_folder(self):
        return self.app.GetDataFolder("IntBoundary")

    @property
    def circuits_folder(self):
        return self.app.GetDataFolder("IntCircuit")

    @property
    def feeders_folder(self):
        return self.app.GetDataFolder("IntFeeder")

    @property
    def versions_folder(self):
        return self.get_unique_obj("*.IntVersionman")

    def get_active_study_case(
        self, error_if_no_active_case: bool = True
    ) -> IntCase | None:
        """Get the currently active study case. Control whether error should be raised if no case is active.

        Args:
            error_if_no_active_case (bool, optional): If True, raise exception if no case is active. If False, return none. Defaults to True.

        Raises:
            powfacpy.PFNoActiveStudyCaseError: When no case is active.

        Returns:
            IntCase: The active study case | None
        """
        case = self.app.GetActiveStudyCase()
        if case or not error_if_no_active_case:
            return case
        else:
            raise powfacpy.exceptions.PFNoActiveStudyCaseError()

    def get_from_study_case(
        self, name: str, if_not_unique: str = "warning", if_no_study_case: str = "error"
    ) -> PFGeneral:
        """Get objects from active study case (similar to PF built-in function 'app.GetFromStudyCase()').

        Additionally, this method prints a warning or raises an exception if there is more than one object found in the study case and if no study case is activated.

        Args:
            name (str): class name of the object (e.g. 'ElmRes'), optionally preceded by an object name without wildcards and a dot (e.g. 'All Calcualations.ElmRes')

            if_not_unique (str, optional): Warn ('warning') or raise exception ('error') if there are more than one objets of class 'class_name'. Defaults to "warning".

            if_no_study_case (str, optional): Warn ('warning') or raise exception ('error') if no study case is active. Defaults to "error".

        Raises:
            powfacpy.PFNoActiveStudyCaseError: No study case activated
            TypeError: More than one object was found

        Returns:
            PFGeneral: Found or created object
        """
        object = self.app.GetFromStudyCase(name)

        if if_no_study_case and not self.app.GetActiveStudyCase():
            if if_no_study_case == "warning":
                warn(
                    "No study case activated. PowerFactory creates object of class_name in tmp folder, outside any study case."
                )
            elif if_no_study_case == "error":
                raise powfacpy.exceptions.PFNoActiveStudyCaseError()

        if if_not_unique and self.is_pf_class(name):
            name = "*." + name
            all_objects_of_this_class = self.get_obj(
                name, parent_folder=object.GetParent(), include_subfolders=False
            )
            if len(all_objects_of_this_class) > 1:
                parent_path = self.get_path_of_object(object.GetParent())
                if if_not_unique == "warning":
                    warn(
                        f"The returned {name} object is not unique in the  study case: '{parent_path}'. Make sure that the correct {name} object is used."
                    )
                if if_not_unique == "error":
                    raise TypeError(
                        f"The returned {name} object is not unique in its folder / in its study case: '{parent_path}'"
                    )
        return object

    def get_results_obj_from_initial_conditions_calc(self) -> ElmRes:
        """Get results object (ElmRes) from the initial conditions calculation object (ComInc).

        This is the results object where results from time domain (RMS/EMT) simulation are written to.

        Returns:
            ElmRes: ElmRes object
        """
        return self.get_from_study_case("ComInc", if_not_unique="error").p_resvar

    def get_events_folder_from_initial_conditions_calc(self) -> IntEvt:
        """Get events folder (IntEvt) from the initial conditions calculation object (ComInc).

        This folder is used for the events in dynamic time domain simulation (RMS/EMT).

        Returns:
            IntEvt: Events folder.
        """
        return self.get_from_study_case("ComInc", if_not_unique="error").p_event

    def get_calc_relevant_obj(
        self,
        obj_str: str,
        condition: Callable = lambda x: True,
        error_if_non_existent=True,
        includeOutOfService: int = 1,
        topoElementsOnly: int = 0,
        bAcSchemes: int = 0,
    ) -> list[PFGeneral]:
        """Wraps the method 'GetCalcRelevantObjects' (see PF scripting reference) and adds optional arguments similar to 'get_obj'.

        Args:
            obj_str (str): name inlcuding class of object(s) (NOT their path)

            condition (Callable, optional): See get_obj. Defaults to lambdax:True.

            error_if_non_existent (bool, optional): See get_obj. Defaults to True.

            From scripting reference:

            includeOutOfService (int, optional): Flag whether to include out of service objects. Defaults to 1.

            topoElementsOnly (int, optional): Flag to filter for topology relevant objects only. Defaults to 0.

            bAcSchemes (int, optional): Flag to include hidden objects in active schemes. Defaults to 0.

        Returns:
            list[PFGeneral]: Found object(s)
        """
        objs = self.app.GetCalcRelevantObjects(
            obj_str, includeOutOfService, topoElementsOnly, bAcSchemes
        )
        if not objs:
            return self._handle_non_existing_obj(
                obj_str, self.get_active_project(), error_if_non_existent
            )
        if condition:
            obj_with_condition = self.get_by_condition(objs, condition)
            if obj_with_condition:
                return obj_with_condition
            else:
                return self._handle_condition_of_obj_not_met(
                    obj_str, self.get_active_project(), error_if_non_existent
                )
        else:
            return objs

    def add_results_variable(
        self,
        obj: PFGeneral | str | list[PFGeneral | str],
        variables: str | list[str],
        results_obj: ElmRes = None,
    ) -> ElmRes:
        """Add variable(s) of 'obj' to the monitored variables in of result object.

        Args:

            obj (PFGeneral | str | list[PFGeneral | str]): PF object or its path

            variables (list[str]): variable names

            results_obj (ElmRes, optional): Results object. Defaults to None (ElmRes from active study case is used).

        Returns:
            ElmRes: the results object
        """
        if not results_obj:
            results_obj = self.get_from_study_case("ElmRes")
        else:
            results_obj = self._handle_single_pf_object_or_path_input(results_obj)
        obj = self._handle_pf_object_or_path_input(obj)
        if isinstance(variables, str):
            variables = [variables]
        for o in obj:
            for var in variables:
                results_obj.AddVariable(o, var)
        results_obj.Load()
        return results_obj

    def add_variable_selection_obj_to_results_obj(
        self,
        name,
        results_obj: ElmRes,
        class_name: str = None,
        variables: list[str] = [],
    ) -> IntMon:
        """Add a varible selection object (IntMon) to a result object (ElmRes).

        Args:
            name (str): Name of IntMon
            results_obj (ElmRes): Results object
            class_name (str, optional): 'classnm' parameter of IntMon. Defaults to None.
            variables (list[str], optional): 'vars' parameter of IntMon. Defaults to [].

        Returns:
            IntMon: varible selection object
        """
        variable_selection_obj: IntMon = self.create_in_folder(
            name + ".IntMon", results_obj
        )
        if class_name:
            variable_selection_obj.classnm = class_name
        if variables:
            variable_selection_obj.vars = variables
        return variable_selection_obj

    def get_first_level_folder(self, folder_type: str) -> PFGeneral:
        """Get folder on first level of PF database.

        Args:
            folder_type (str): The folder of the active user ('user') or the global library ('global library') can be accessed.

        Raises:
            TypeError: Invalid folder_tpe input

        Returns:
            PFGeneral: first level folder
        """
        if folder_type == "user":
            return self.app.GetCurrentUser()
        elif folder_type == "global library":
            return self.app.GetGlobalLibrary()
        else:
            raise TypeError(
                f"The first level folder {folder_type} is not valid. Use one of these: 'user', 'global library'."
            )

    def get_active_user_folder(self) -> IntUser:
        """Get folder of active user."""
        return self.app.GetCurrentUser()

    def get_active_networks(self, error_if_no_network_is_active: bool = True) -> ElmNet:
        """Get active networks/grids."""
        grids = self.app.GetCalcRelevantObjects(
            ".ElmNet"
        )  # This also returns the summary grid in the study case
        # Delete the summary grid which is in the study case
        grids[:] = [
            grid for grid in grids if not grid.GetParent().GetClassName() == "IntCase"
        ]
        if error_if_no_network_is_active and not grids:
            raise powfacpy.exceptions.PFNotActiveError("a network (ElmNet).")
        return grids

    def activate_study_case(self, path: str) -> IntCase:
        """Activate study case under path."""
        study_case = self.get_unique_obj(path, include_subfolders=False)
        study_case.Activate()
        return study_case

    def clear_elmres_from_objects_with_status_deleted(self, results_obj: ElmRes = None):
        """Deletes all objects from a results object (ElmRes) that have the
        status deleted (i.e. attribute 'obj_id' is deleted).
        """
        if not results_obj:
            results_obj = self.get_from_study_case("ElmRes")
        obj_in_elmres = results_obj.GetContents("*")
        for o in obj_in_elmres:
            obj_id = o.obj_id
            if obj_id.IsDeleted():
                o.Delete()

    def clear_elmres(self, results_obj: ElmRes = None):
        """Clear all results variables from results object (ElmRes).

        Args:
            results_obj (ElmRes, optional): Results object. Defaults to None (get elmres from study case).
        """
        if not results_obj:
            results_obj = self.get_from_study_case("ElmRes")
        self.clear_folder(results_obj)

    def get_parameter_value_string(self, parameters: dict, delimiter=" ") -> str:
        """Get string with parameters and their values.

        Args:
            parameters (dict): parameters (keys) and values (values)
            Example: {'P': 'user\\project\\path\\to\\object\\m:Psum:bus1'}

            delimiter (str, optional): Delimiter between parameter value pairs. Defaults to " ".

        Returns:
            str: parameter value string (e.g. 'P = 2.5')
        """
        param_value_string = ""
        for parname, path_with_par in parameters.items():
            value = self.get_attr_by_path(path_with_par)
            param_value_string += parname + "=" + str(value) + delimiter
        return param_value_string[: -len(delimiter)]  # omit last delimiter

    def create_comtrade_obj(
        self, file_path: str, parent_folder: Union[PFGeneral, str] = None
    ) -> IntComtrade:
        """Add an IntComtrade that refers to file_path (*.cfg).
        The objects are stored in a folder "Comtrade" in the currently active
        study case, unless a parent_folder is given. A new object is only
        created if there exists no object yet that points to the same file
        ('f_name' attribute is the file path). The file name is used for the
        new object name (without the .cfg ending).
        """
        if parent_folder:
            parent_folder = self._handle_single_pf_object_or_path_input(parent_folder)
        else:
            parent_folder = self.app.GetFromStudyCase("Comtrade.IntFolder")

        intcomtrade: IntComtrade = self.get_obj(
            "*.IntComtrade",
            parent_folder=parent_folder,
            condition=lambda x: getattr(x, "f_name") == file_path,
            error_if_non_existent=False,
        )
        if not intcomtrade:
            _, file_name = os_path.split(file_path)
            intcomtrade = self.create_in_folder(
                file_name.replace(".cfg", "") + ".IntComtrade",
                parent_folder,
                overwrite=False,
                use_existing=False,
            )
            intcomtrade.f_name = file_path
        else:
            intcomtrade = intcomtrade[0]
        # intcomtrade.Load() probably not required
        return intcomtrade

    def create_zone(
        self,
        name: str,
        elms: list[PFGeneral],
        color: int = 2,
        parent_folder: str | PFGeneral = None,
    ) -> ElmZone:
        if not parent_folder:
            parent_folder = self.zones_folder
        elmzone: ElmZone = self.create_in_folder(name + ".ElmZone", parent_folder)
        elmzone.icolor = color
        for elm in elms:
            elm.cpZone = elmzone
        return elmzone

    def create_variation(
        self,
        name: str,
        parent_folder: str | PFGeneral = None,
        name_expansion_stage: str = "Expansion Stage",
        activationTime: int = 0,
        activate: int = 1,
    ) -> IntScheme:
        """Create variation (including one expansion stage).

        Args:
            name (str): Name of variation
            parent_folder (str | PFGeneral, optional): Parent folder where variation is created. Defaults to None (i.e. variations folder).
            name_expansion_stage (str, optional): Name of. Defaults to "Expansion Stage".
            activationTime (int, optional): UTC time
            activate (int, optional): If 1, expansion stage is activate. If 0, expansion stage is not activated. Defaults to 1.

        Returns:
            IntScheme: The created variation object
        """
        if not parent_folder:
            parent_folder = self.variations_folder
        variation = self.create_in_folder(name + ".IntScheme", parent_folder)

        variation.NewStage(name_expansion_stage, activationTime, activate)
        return variation

    def execute_load_flow(self, params: dict = {}) -> int:
        comldf: ComLdf = self.get_from_study_case("ComLdf")
        self.set_attr(comldf, params=params)
        return comldf.Execute()

    def mark_in_graphics(
        self, elms: list[PFGeneral] | PFGeneral, searchOpenedDiagramsOnly: int = 0
    ) -> None:
        if not isinstance(elms, list):
            elms = list(elms)
        self.app.MarkInGraphics(elms, searchOpenedDiagramsOnly)

    def _handle_possible_attribute_not_set_error(
        self, possibly_not_secified_attr: str, attribute_description: str, error_message
    ):
        """Handles the exception if in a method call (of this class) an AttributeError is raised because of an attribute of this class is not set (i.e. is None).

        Example:
            A method uses self.active_case and active_case is None.
        """
        if not possibly_not_secified_attr:
            raise powfacpy.exceptions.PFAttributeNotSetError(attribute_description)
        else:
            raise AttributeError(error_message)

    def get_project_version(self, version_name: str) -> IntVersion | None:
        """Get (previous) version of project.

        Args:
            version_name (str): Name (loc_name) of version

        Returns:
            IntVersion | None: Version object
        """
        version = self.get_by_condition(
            self._folder.GetVersions(), lambda x: x.loc_name == version_name
        )
        if version:
            return version[0]

    def create_project_version(self, version_name: str, overwrite: bool = True) -> None:
        """Create a version of current state of the project.

        Uses 'CreateVersion'. New version will be added to top level versions folder of project.

        Args:
            version_name (str): Name (loc_name) of version

            overwrite (bool, optional): Overwrite existing version with same name. Defaults to True.
        """
        version = self.get_project_version(version_name)
        if version and overwrite:
            version.Delete()
        self.app.WriteChangesToDb()
        self._folder.CreateVersion(version_name)

    def rollback_project_to_previous_version(self, version_name: str) -> None:
        """Rollback to previous project version (IntVersion in versions folder).

        Args:
            version_name (str): Name (loc_name) of version.
        """
        active_project = self._folder
        version = self.get_project_version(version_name)
        try:
            active_project.Deactivate()
            version.Rollback()
        finally:
            active_project.Activate()

    def import_project(
        self,
        file_path: str,
        target_folder_in_active_user: str | PFGeneral | None = None,
        keep_current_project_activated: bool = True,
    ) -> IntPrj:
        """Import a project (.pfd file)

        Args:
            file_path (str): Windows path
            target_folder_in_active_user (str | PFGeneral | None, optional): Target folder for project import in active user. Defaults to None.
            keep_current_project_activated (bool, optional): If True, the initial project and study case remain active.If False, the imported project will be active after import. Defaults to True.

        Returns:
            IntPrj: Imported project
        """
        try:
            if keep_current_project_activated:
                initial_project = self._folder
                initial_study_case = self.get_active_study_case()
            pfd_import: ComPfdimport = self.get_from_study_case("ComPfdimport")
            if not file_path[-4:] == ".pfd":
                file_path += ".pfd"
            pfd_import.g_file = file_path
            if target_folder_in_active_user:
                if isinstance(target_folder_in_active_user, str):
                    pfd_import.g_target = self.get_unique_obj(
                        target_folder_in_active_user,
                        parent_folder=self.get_active_user_folder(),
                    )
                else:
                    pfd_import.g_target = target_folder_in_active_user
            else:
                pfd_import.g_target = self.get_active_user_folder()
            pfd_import.Execute()
            imported_project: IntPrj = self.app.GetActiveProject()
        finally:
            if keep_current_project_activated:
                imported_project.Deactivate()
                initial_project.Activate()
                initial_study_case.Activate()
        return imported_project

    @staticmethod
    def replace_headers_of_csv_file_with_number_of_colums(file_path: str) -> int:
        """Replaces the first row (headers) of a csv file with its number of
        columns. This is needed for import of csv files to PF using ElmFile.
        """
        with (
            open(file_path + ".csv") as read_file,
            open(file_path + ".temp", "w") as write_file,
        ):
            columns_of_first_row = read_file.readline().split(",")
            if columns_of_first_row[-1] == "\n":
                columns_of_first_row = columns_of_first_row[:-1]
            # Minus one because first column is time and should not be counted.
            number_of_columns = len(columns_of_first_row) - 1
            write_file.write(str(number_of_columns) + "\n")
            row = read_file.readline()
            while row:
                write_file.write(row)
                row = read_file.readline()
        replace(file_path + ".temp", file_path + ".csv")
        return number_of_columns

    @staticmethod
    def insert_row_with_number_of_columns_in_csv_file(file_path: str) -> int:
        """Gets the number of columns of the first row in a csv file and
        inserts a row (first row) with this number in the first column.
        This is needed for ElmFile to read csv files.
        """
        with (
            open(file_path + ".csv") as read_file,
            open(file_path + ".temp", "w") as write_file,
        ):
            first_row = read_file.readline()
            columns_of_first_row = first_row.split(",")
            if columns_of_first_row[-1] == "\n":
                columns_of_first_row = columns_of_first_row[:-1]
            # Minus one because first column is time and should not be counted.
            number_of_columns = len(columns_of_first_row) - 1
            write_file.write(str(number_of_columns) + "\n")
            write_file.write(first_row)
            row = read_file.readline()
            while row:
                write_file.write(row)
                row = read_file.readline()
        replace(file_path + ".temp", file_path + ".csv")
        return number_of_columns


class PFBaseInterface(PFActiveProject):
    """Please use the class PFActiveProject instead."""

    def __init__(self, pf_app: PFApp):
        warn(
            f"{self.__class__.__name__} will be deprecated. Please use the class PFActiveProject instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(pf_app)
